from __future__ import annotations

from dataclasses import dataclass

from chik.util.ints import uint8, uint32
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class FeeStatBackup(Streamable):
    type: str
    tx_ct_avg: list[str]
    confirmed_average: list[list[str]]
    failed_average: list[list[str]]
    m_fee_rate_avg: list[str]


@streamable
@dataclass(frozen=True)
class FeeTrackerBackup(Streamable):
    fee_estimator_version: uint8
    first_recorded_height: uint32
    latest_seen_height: uint32
    stats: list[FeeStatBackup]



from __future__ import annotations

import time
from pathlib import Path
from secrets import token_bytes
from typing import Any, Optional, Union

import pytest
from typing_extensions import override

from chik._tests.cmds.cmd_test_utils import TestRpcClients, TestWalletRpcClient, run_cli_command_and_assert
from chik._tests.cmds.wallet.test_consts import FINGERPRINT_ARG, STD_TX, STD_UTX
from chik.rpc.wallet_request_types import (
    CreateNewDAOWalletResponse,
    DAOAddFundsToTreasuryResponse,
    DAOCloseProposalResponse,
    DAOCreateProposalResponse,
    DAOExitLockupResponse,
    DAOFreeCoinsFromFinishedProposalsResponse,
    DAOSendToLockupResponse,
    DAOVoteOnProposalResponse,
)
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import encode_puzzle_hash
from chik.util.ints import uint8, uint32, uint64
from chik.wallet.conditions import ConditionValidTimes, parse_timelock_info
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.util.transaction_type import TransactionType
from chik.wallet.util.tx_config import TXConfig
from chik.wallet.util.wallet_types import WalletType

# DAO Commands


def test_dao_create(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class DAOCreateRpcClient(TestWalletRpcClient):
        async def create_new_dao_wallet(
            self,
            mode: str,
            tx_config: TXConfig,
            dao_rules: Optional[dict[str, uint64]] = None,
            amount_of_cats: Optional[uint64] = None,
            treasury_id: Optional[bytes32] = None,
            filter_amount: uint64 = uint64(1),
            name: Optional[str] = None,
            fee: uint64 = uint64(0),
            fee_for_cat: uint64 = uint64(0),
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> CreateNewDAOWalletResponse:
            if not treasury_id:
                treasury_id = bytes32(token_bytes(32))
            return CreateNewDAOWalletResponse.from_json_dict(
                {
                    "success": True,
                    "transactions": [STD_TX.to_json_dict()],
                    "unsigned_transactions": [STD_UTX.to_json_dict()],
                    "type": WalletType.DAO,
                    "wallet_id": 2,
                    "treasury_id": treasury_id,
                    "cat_wallet_id": 3,
                    "dao_cat_wallet_id": 4,
                }
            )

    inst_rpc_client = DAOCreateRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = [
        "dao",
        "create",
        FINGERPRINT_ARG,
        "-n test",
        "--attendance-required",
        "1000",
        "--cat-amount",
        "100000",
        "-m0.1",
        "--reuse",
    ]
    # these are various things that should be in the output
    assert_list = ["Successfully created DAO Wallet", "DAO Wallet ID: 2", "CAT Wallet ID: 3", "DAOCAT Wallet ID: 4"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)

    # Check command raises if proposal minimum is even
    odd_pm_command_args = [
        "dao",
        "create",
        FINGERPRINT_ARG,
        "-n test",
        "--attendance-required",
        "1000",
        "--cat-amount",
        "100000",
        "--proposal-minimum",
        "10",
        "-m0.1",
        "--reuse",
    ]
    extra_assert_list = [
        "Adding 1 mojo to proposal minimum amount",
    ]
    run_cli_command_and_assert(capsys, root_dir, odd_pm_command_args, extra_assert_list)

    # Add wallet for existing DAO
    add_command_args = [
        "dao",
        "add",
        FINGERPRINT_ARG,
        "-n test",
        "-t",
        bytes32(token_bytes(32)).hex(),
        "--filter-amount",
        "1",
    ]
    run_cli_command_and_assert(capsys, root_dir, add_command_args, assert_list)


def test_dao_treasury(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    class DAOCreateRpcClient(TestWalletRpcClient):
        async def dao_get_treasury_id(
            self,
            wallet_id: int,
        ) -> dict[str, str]:
            return {"treasury_id": "0xCAFEF00D"}

        async def dao_get_treasury_balance(self, wallet_id: int) -> dict[str, Union[str, bool, dict[str, int]]]:
            if wallet_id == 2:
                return {"success": True, "balances": {"xck": 1000000000000, "0xCAFEF00D": 10000000}}
            else:
                return {"success": True, "balances": {}}

        async def dao_add_funds_to_treasury(
            self,
            wallet_id: int,
            funding_wallet_id: int,
            amount: uint64,
            tx_config: TXConfig,
            fee: uint64 = uint64(0),
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOAddFundsToTreasuryResponse:
            return DAOAddFundsToTreasuryResponse([STD_UTX], [STD_TX], STD_TX.name, STD_TX)

        async def dao_get_rules(
            self,
            wallet_id: int,
        ) -> dict[str, dict[str, int]]:
            return {"rules": {"proposal_minimum": 100}}

        @override
        async def get_transaction(self, transaction_id: bytes32) -> TransactionRecord:
            return TransactionRecord(
                confirmed_at_height=uint32(0),
                created_at_time=uint64(int(time.time())),
                to_puzzle_hash=bytes32(b"2" * 32),
                amount=uint64(10),
                fee_amount=uint64(1),
                confirmed=True,
                sent=uint32(10),
                spend_bundle=None,
                additions=[],
                removals=[],
                wallet_id=uint32(1),
                sent_to=[("peer1", uint8(1), None)],
                trade_id=None,
                type=uint32(TransactionType.INCOMING_TX.value),
                name=bytes32(token_bytes()),
                memos=[],
                valid_times=parse_timelock_info(tuple()),
            )

    inst_rpc_client = DAOCreateRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client

    get_id_args = ["dao", "get_id", FINGERPRINT_ARG, "-i 2"]
    get_id_asserts = ["Treasury ID: 0xCAFEF00D"]
    run_cli_command_and_assert(capsys, root_dir, get_id_args, get_id_asserts)

    get_balance_args = ["dao", "balance", FINGERPRINT_ARG, "-i 2"]
    get_balance_asserts = ["XCK: 1.0", "0xCAFEF00D: 10000.0"]
    run_cli_command_and_assert(capsys, root_dir, get_balance_args, get_balance_asserts)

    no_balance_args = ["dao", "balance", FINGERPRINT_ARG, "-i 3"]
    no_balance_asserts = ["The DAO treasury currently has no funds"]
    run_cli_command_and_assert(capsys, root_dir, no_balance_args, no_balance_asserts)

    add_funds_args = ["dao", "add_funds", FINGERPRINT_ARG, "-i 2", "-w 1", "-a", "10", "-m 0.1", "--reuse"]
    add_funds_asserts = [
        "Transaction submitted to nodes",
    ]
    run_cli_command_and_assert(capsys, root_dir, add_funds_args, add_funds_asserts)

    rules_args = ["dao", "rules", FINGERPRINT_ARG, "-i 2"]
    rules_asserts = "proposal_minimum: 100"
    run_cli_command_and_assert(capsys, root_dir, rules_args, rules_asserts)


def test_dao_proposals(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class DAOCreateRpcClient(TestWalletRpcClient):
        async def dao_get_proposals(
            self,
            wallet_id: int,
            include_closed: bool = True,
        ) -> dict[str, Union[bool, int, list[Any]]]:
            proposal = {
                "proposal_id": "0xCAFEF00D",
                "amount_voted": uint64(10),
                "yes_votes": uint64(10),
                "passed": True,
                "closed": True,
            }
            proposal_2 = {
                "proposal_id": "0xFEEDBEEF",
                "amount_voted": uint64(120),
                "yes_votes": uint64(100),
                "passed": True,
                "closed": False,
            }
            return {
                "success": True,
                "proposals": [proposal, proposal_2],
                "proposal_timelock": 5,
                "soft_close_length": 10,
            }

        async def dao_parse_proposal(
            self,
            wallet_id: int,
            proposal_id: str,
        ) -> dict[str, Union[bool, dict[str, Any]]]:
            if proposal_id == "0xCAFEF00D":
                puzhash = bytes32(b"1" * 32).hex()
                asset_id = bytes32(b"2" * 32).hex()
                proposal_details: dict[str, Any] = {
                    "proposal_type": "s",
                    "xck_conditions": [{"puzzle_hash": puzhash, "amount": 100}],
                    "asset_conditions": [
                        {"asset_id": asset_id, "conditions": [{"puzzle_hash": puzhash, "amount": 123}]}
                    ],
                }
            elif proposal_id == "0xFEEDBEEF":
                proposal_details = {
                    "proposal_type": "u",
                    "dao_rules": {
                        "proposal_timelock": 10,
                        "soft_close_length": 50,
                    },
                }
            else:
                proposal_details = {
                    "proposal_type": "s",
                    "mint_amount": 1000,
                    "new_cat_puzhash": bytes32(b"x" * 32).hex(),
                }
            proposal_state = {
                "state": {
                    "passed": False,
                    "closable": False,
                    "closed": False,
                    "total_votes_needed": 10,
                    "yes_votes_needed": 20,
                    "blocks_needed": 30,
                }
            }
            proposal_dict = {**proposal_state, **proposal_details}
            return {"success": True, "proposal_dictionary": proposal_dict}

        async def dao_vote_on_proposal(
            self,
            wallet_id: int,
            proposal_id: str,
            vote_amount: int,
            tx_config: TXConfig,
            is_yes_vote: bool,
            fee: uint64 = uint64(0),
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOVoteOnProposalResponse:
            return DAOVoteOnProposalResponse([STD_UTX], [STD_TX], STD_TX.name, STD_TX)

        async def dao_close_proposal(
            self,
            wallet_id: int,
            proposal_id: str,
            tx_config: TXConfig,
            fee: uint64 = uint64(0),
            self_destruct: bool = False,
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOCloseProposalResponse:
            return DAOCloseProposalResponse([STD_UTX], [STD_TX], STD_TX.name, STD_TX)

        async def dao_create_proposal(
            self,
            wallet_id: int,
            proposal_type: str,
            tx_config: TXConfig,
            additions: Optional[list[dict[str, Any]]] = None,
            amount: Optional[uint64] = None,
            inner_address: Optional[str] = None,
            asset_id: Optional[str] = None,
            cat_target_address: Optional[str] = None,
            vote_amount: Optional[int] = None,
            new_dao_rules: Optional[dict[str, uint64]] = None,
            fee: uint64 = uint64(0),
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOCreateProposalResponse:
            return DAOCreateProposalResponse([STD_UTX], [STD_TX], bytes32.zeros, STD_TX.name, STD_TX)

        async def get_wallets(self, wallet_type: Optional[WalletType] = None) -> list[dict[str, Union[str, int]]]:
            return [{"id": 1, "type": 0}, {"id": 2, "type": 14}]

        @override
        async def get_transaction(self, transaction_id: bytes32) -> TransactionRecord:
            return TransactionRecord(
                confirmed_at_height=uint32(0),
                created_at_time=uint64(int(time.time())),
                to_puzzle_hash=bytes32(b"2" * 32),
                amount=uint64(10),
                fee_amount=uint64(1),
                confirmed=True,
                sent=uint32(10),
                spend_bundle=None,
                additions=[],
                removals=[],
                wallet_id=uint32(1),
                sent_to=[("peer1", uint8(1), None)],
                trade_id=None,
                type=uint32(TransactionType.INCOMING_TX.value),
                name=bytes32(b"x" * 32),
                memos=[],
                valid_times=parse_timelock_info(tuple()),
            )

    # List all proposals
    inst_rpc_client = DAOCreateRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    list_args = ["dao", "list_proposals", FINGERPRINT_ARG, "-i 2"]
    # these are various things that should be in the output
    list_asserts = [
        "Proposal ID: 0xCAFEF00D",
        "Status: OPEN",
        "Votes for: 10",
        "Votes against: 0",
        "Proposal ID: 0xFEEDBEEF",
        "Status: CLOSED",
        "Votes for: 100",
        "Votes against: 20",
        "Proposals have 10 blocks of soft close time.",
    ]
    run_cli_command_and_assert(capsys, root_dir, list_args, list_asserts)

    # Show details of specific proposal
    parse_spend_args = ["dao", "show_proposal", FINGERPRINT_ARG, "-i 2", "-p", "0xCAFEF00D"]
    address = encode_puzzle_hash(bytes32(b"1" * 32), "xck")
    asset_id = bytes32(b"2" * 32).hex()
    parse_spend_asserts = [
        "Type: SPEND",
        "Status: OPEN",
        "Passed: False",
        "Closable: False",
        "Total votes needed: 10",
        "Yes votes needed: 20",
        "Blocks remaining: 30",
        "Proposal XCK Conditions",
        f"Address: {address}",
        "Amount: 100",
        "Proposal asset Conditions",
        f"Asset ID: {asset_id}",
        f"Address: {address}",
        "Amount: 123",
    ]
    run_cli_command_and_assert(capsys, root_dir, parse_spend_args, parse_spend_asserts)

    parse_update_args = ["dao", "show_proposal", FINGERPRINT_ARG, "-i2", "-p", "0xFEEDBEEF"]
    parse_update_asserts = [
        "Type: UPDATE",
        "proposal_timelock: 10",
        "soft_close_length: 50",
    ]
    run_cli_command_and_assert(capsys, root_dir, parse_update_args, parse_update_asserts)

    parse_mint_args = ["dao", "show_proposal", FINGERPRINT_ARG, "-i2", "-p", "0xDABBAD00"]
    parse_mint_asserts = [
        "Type: MINT",
        "Amount of CAT to mint: 1000",
        f"Address: {encode_puzzle_hash(bytes32(b'x' * 32), 'xck')}",
    ]
    run_cli_command_and_assert(capsys, root_dir, parse_mint_args, parse_mint_asserts)

    # Vote on a proposal
    vote_args = ["dao", "vote", FINGERPRINT_ARG, "-i 2", "-p", "0xFEEDBEEF", "-a", "1000", "-n", "-m 0.1", "--reuse"]
    vote_asserts = ["Transaction submitted to nodes"]
    run_cli_command_and_assert(capsys, root_dir, vote_args, vote_asserts)

    # Close a proposal
    close_args = ["dao", "close_proposal", FINGERPRINT_ARG, "-i 2", "-p", "0xFEEDBEEF", "-d", "-m 0.1", "--reuse"]
    close_asserts = ["Transaction submitted to nodes"]
    run_cli_command_and_assert(capsys, root_dir, close_args, close_asserts)

    # Create a spend proposal
    address = encode_puzzle_hash(bytes32(b"x" * 32), "xck")
    spend_args = [
        "dao",
        "create_proposal",
        "spend",
        FINGERPRINT_ARG,
        "-i 2",
        "-t",
        address,
        "-a",
        "10",
        "-v",
        "1000",
        "--asset-id",
        "0xFEEDBEEF",
        "-m 0.1",
        "--reuse",
    ]
    proposal_asserts = ["Successfully created proposal", f"Proposal ID: {bytes32.zeros.hex()}"]
    run_cli_command_and_assert(capsys, root_dir, spend_args, proposal_asserts)

    bad_spend_args = [
        "dao",
        "create_proposal",
        "spend",
        FINGERPRINT_ARG,
        "-i 2",
        "-t",
        address,
        "-v",
        "1000",
        "--asset-id",
        "0xFEEDBEEF",
        "-m 0.1",
        "--reuse",
    ]
    proposal_asserts = ["Successfully created proposal", f"Proposal ID: {bytes32.zeros.hex()}"]
    with pytest.raises(ValueError) as e_info:
        run_cli_command_and_assert(capsys, root_dir, bad_spend_args, proposal_asserts)
    assert e_info.value.args[0] == "Must include a json specification or an address / amount pair."

    # Create an update proposal
    update_args = [
        "dao",
        "create_proposal",
        "update",
        FINGERPRINT_ARG,
        "-i 2",
        "-v",
        "1000",
        "--proposal-timelock",
        "4",
        "-m 0.1",
        "--reuse",
    ]
    run_cli_command_and_assert(capsys, root_dir, update_args, proposal_asserts)

    # Create a mint proposal
    mint_args = [
        "dao",
        "create_proposal",
        "mint",
        FINGERPRINT_ARG,
        "-i 2",
        "-v",
        "1000",
        "-a",
        "100",
        "-t",
        address,
        "-m 0.1",
        "--reuse",
    ]
    run_cli_command_and_assert(capsys, root_dir, mint_args, proposal_asserts)


def test_dao_cats(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class DAOCreateRpcClient(TestWalletRpcClient):
        async def dao_send_to_lockup(
            self,
            wallet_id: int,
            amount: uint64,
            tx_config: TXConfig,
            fee: uint64 = uint64(0),
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOSendToLockupResponse:
            return DAOSendToLockupResponse([STD_UTX], [STD_TX], STD_TX.name, [STD_TX])

        async def dao_free_coins_from_finished_proposals(
            self,
            wallet_id: int,
            tx_config: TXConfig,
            fee: uint64 = uint64(0),
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOFreeCoinsFromFinishedProposalsResponse:
            return DAOFreeCoinsFromFinishedProposalsResponse([STD_UTX], [STD_TX], STD_TX.name, STD_TX)

        async def dao_exit_lockup(
            self,
            wallet_id: int,
            tx_config: TXConfig,
            coins: Optional[list[dict[str, Union[str, int]]]] = None,
            fee: uint64 = uint64(0),
            reuse_puzhash: Optional[bool] = None,
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> DAOExitLockupResponse:
            return DAOExitLockupResponse([STD_UTX], [STD_TX], STD_TX.name, STD_TX)

        @override
        async def get_transaction(self, transaction_id: bytes32) -> TransactionRecord:
            return TransactionRecord(
                confirmed_at_height=uint32(0),
                created_at_time=uint64(int(time.time())),
                to_puzzle_hash=bytes32(b"2" * 32),
                amount=uint64(10),
                fee_amount=uint64(1),
                confirmed=True,
                sent=uint32(10),
                spend_bundle=None,
                additions=[],
                removals=[],
                wallet_id=uint32(1),
                sent_to=[("peer1", uint8(1), None)],
                trade_id=None,
                type=uint32(TransactionType.INCOMING_TX.value),
                name=bytes32(b"x" * 32),
                memos=[],
                valid_times=parse_timelock_info(tuple()),
            )

    inst_rpc_client = DAOCreateRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    lockup_args = ["dao", "lockup_coins", FINGERPRINT_ARG, "-i 2", "-a", "1000", "-m 0.1", "--reuse"]
    lockup_asserts = ["Transaction submitted to nodes"]
    run_cli_command_and_assert(capsys, root_dir, lockup_args, lockup_asserts)

    release_args = ["dao", "release_coins", FINGERPRINT_ARG, "-i 2", "-m 0.1", "--reuse"]
    # tx_id = bytes32(b"x" * 32).hex()
    release_asserts = ["Transaction submitted to nodes"]
    run_cli_command_and_assert(capsys, root_dir, release_args, release_asserts)

    exit_args = ["dao", "exit_lockup", FINGERPRINT_ARG, "-i 2", "-m 0.1", "--reuse"]
    exit_asserts = ["Transaction submitted to nodes"]
    run_cli_command_and_assert(capsys, root_dir, exit_args, exit_asserts)



from __future__ import annotations

from enum import Enum
from typing import Optional

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.wallet.cat_wallet.cat_outer_puzzle import CATOuterPuzzle
from chik.wallet.driver_protocol import DriverProtocol
from chik.wallet.nft_wallet.metadata_outer_puzzle import MetadataOuterPuzzle
from chik.wallet.nft_wallet.ownership_outer_puzzle import OwnershipOuterPuzzle
from chik.wallet.nft_wallet.singleton_outer_puzzle import SingletonOuterPuzzle
from chik.wallet.nft_wallet.transfer_program_puzzle import TransferProgramPuzzle
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.uncurried_puzzle import UncurriedPuzzle
from chik.wallet.vc_wallet.cr_outer_puzzle import CROuterPuzzle

"""
This file provides a central location for acquiring drivers for outer puzzles like CATs, NFTs, etc.

A driver for a puzzle must include the following functions:
  - match(self, puzzle: UncurriedPuzzle) -> Optional[PuzzleInfo]
    - Given a puzzle reveal, return a PuzzleInfo object that can be used to reconstruct it later
  - get_inner_puzzle(self, constructor: PuzzleInfo, puzzle_reveal: UncurriedPuzzle) -> Optional[Program]:
    - Given a PuzzleInfo object and a puzzle reveal, pull out this outer puzzle's inner puzzle
  - asset_id(self, constructor: PuzzleInfo) -> Optional[bytes32]
    - Given a PuzzleInfo object, generate a 32 byte ID for use in dictionaries, etc.
  - construct(self, constructor: PuzzleInfo, inner_puzzle: Program) -> Program
    - Given a PuzzleInfo object and an innermost puzzle, construct a puzzle reveal for a coin spend
  - solve(self, constructor: PuzzleInfo, solver: Solver, inner_puzzle: Program, inner_solution: Program) -> Program
    - Given a PuzzleInfo object, a Solver object, and an innermost puzzle and its solution return a solution for a spend
    - The "Solver" object can contain any dictionary, it's up to the driver to enforce the needed elements of the API
    - Some classes that wish to integrate with a driver may not have access to all of the info it needs so the driver
      needs to raise errors appropriately
"""


class AssetType(Enum):
    CAT = "CAT"
    SINGLETON = "singleton"
    METADATA = "metadata"
    OWNERSHIP = "ownership"
    ROYALTY_TRANSFER_PROGRAM = "royalty transfer program"
    CR = "credential restricted"


def match_puzzle(puzzle: UncurriedPuzzle) -> Optional[PuzzleInfo]:
    for driver in driver_lookup.values():
        potential_info: Optional[PuzzleInfo] = driver.match(puzzle)
        if potential_info is not None:
            return potential_info
    return None


def construct_puzzle(constructor: PuzzleInfo, inner_puzzle: Program) -> Program:
    return driver_lookup[AssetType(constructor.type())].construct(constructor, inner_puzzle)


def solve_puzzle(constructor: PuzzleInfo, solver: Solver, inner_puzzle: Program, inner_solution: Program) -> Program:
    return driver_lookup[AssetType(constructor.type())].solve(constructor, solver, inner_puzzle, inner_solution)


def get_inner_puzzle(constructor: PuzzleInfo, puzzle_reveal: UncurriedPuzzle) -> Optional[Program]:
    return driver_lookup[AssetType(constructor.type())].get_inner_puzzle(constructor, puzzle_reveal)


def get_inner_solution(constructor: PuzzleInfo, solution: Program) -> Optional[Program]:
    return driver_lookup[AssetType(constructor.type())].get_inner_solution(constructor, solution)


def create_asset_id(constructor: PuzzleInfo) -> Optional[bytes32]:
    return driver_lookup[AssetType(constructor.type())].asset_id(constructor)


function_args = (match_puzzle, construct_puzzle, solve_puzzle, get_inner_puzzle, get_inner_solution)

driver_lookup: dict[AssetType, DriverProtocol] = {
    AssetType.CAT: CATOuterPuzzle(*function_args),
    AssetType.SINGLETON: SingletonOuterPuzzle(*function_args),
    AssetType.METADATA: MetadataOuterPuzzle(*function_args),
    AssetType.OWNERSHIP: OwnershipOuterPuzzle(*function_args),
    AssetType.ROYALTY_TRANSFER_PROGRAM: TransferProgramPuzzle(*function_args),
    AssetType.CR: CROuterPuzzle(*function_args),
}



from __future__ import annotations

import logging

import pytest
from klvm.casts import int_to_bytes

from chik._tests.blockchain.blockchain_test_utils import _validate_and_add_block
from chik._tests.util.generator_tools_testing import run_and_get_removals_and_additions
from chik.consensus.blockchain import AddBlockResult
from chik.full_node.full_node_api import FullNodeAPI
from chik.protocols import wallet_protocol
from chik.server.server import ChikServer
from chik.simulator.add_blocks_in_batches import add_blocks_in_batches
from chik.simulator.block_tools import BlockTools, test_constants
from chik.simulator.wallet_tools import WalletTool
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.condition_with_args import ConditionWithArgs
from chik.types.spend_bundle import SpendBundle, estimate_fees
from chik.util.errors import Err
from chik.util.ints import uint32, uint64
from chik.wallet.conditions import AssertCoinAnnouncement, AssertPuzzleAnnouncement

BURN_PUZZLE_HASH = bytes32(b"0" * 32)

WALLET_A = WalletTool(test_constants)
WALLET_A_PUZZLE_HASHES = [WALLET_A.get_new_puzzlehash() for _ in range(5)]

log = logging.getLogger(__name__)


class TestBlockchainTransactions:
    @pytest.mark.anyio
    async def test_basic_blockchain_tx(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[2]
        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin

        assert spend_coin is not None
        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_puzzlehash, spend_coin)

        assert spend_bundle is not None
        tx: wallet_protocol.SendTransaction = wallet_protocol.SendTransaction(spend_bundle)

        await full_node_api_1.send_transaction(tx)

        sb = full_node_1.mempool_manager.get_spendbundle(spend_bundle.name())
        assert sb == spend_bundle

        last_block = blocks[-1]
        result = await full_node_1.mempool_manager.create_bundle_from_mempool(
            last_block.header_hash, full_node_1.coin_store.get_unspent_lineage_info_for_puzzle_hash
        )
        assert result is not None
        next_spendbundle, _ = result

        new_blocks = bt.get_consecutive_blocks(
            1,
            block_list_input=blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=next_spendbundle,
            guarantee_transaction_block=True,
        )

        next_block = new_blocks[-1]
        await full_node_1.add_block(next_block)

        blockchain_peak = full_node_1.blockchain.get_peak()
        assert blockchain_peak is not None
        assert next_block.header_hash == blockchain_peak.header_hash

        added_coins = next_spendbundle.additions()

        # Two coins are added, main spend and change
        assert len(added_coins) == 2
        for coin in added_coins:
            unspent = await full_node_1.coin_store.get_coin_record(coin.name())
            assert unspent is not None
            assert not unspent.spent
            assert not unspent.coinbase

    @pytest.mark.anyio
    async def test_validate_blockchain_with_double_spend(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 5
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[2]
        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None

        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_puzzlehash, spend_coin)
        spend_bundle_double = wallet_a.generate_signed_transaction(uint64(1001), receiver_puzzlehash, spend_coin)

        block_spendbundle = SpendBundle.aggregate([spend_bundle, spend_bundle_double])

        new_blocks = bt.get_consecutive_blocks(
            1,
            block_list_input=blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block_spendbundle,
            guarantee_transaction_block=True,
        )

        next_block = new_blocks[-1]
        await _validate_and_add_block(full_node_1.blockchain, next_block, expected_error=Err.DOUBLE_SPEND)

    @pytest.mark.anyio
    async def test_validate_blockchain_duplicate_output(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 3
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[2]

        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None

        spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin, additional_outputs=[(receiver_puzzlehash, 1000)]
        )

        new_blocks = bt.get_consecutive_blocks(
            1,
            block_list_input=blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
            guarantee_transaction_block=True,
        )

        next_block = new_blocks[-1]
        await _validate_and_add_block(full_node_1.blockchain, next_block, expected_error=Err.DUPLICATE_OUTPUT)

    @pytest.mark.anyio
    async def test_validate_blockchain_with_reorg_double_spend(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[2]

        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None

        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_puzzlehash, spend_coin)

        blocks_spend = bt.get_consecutive_blocks(
            1,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
        )
        # Move chain to height 10, with a spend at height 10
        await add_blocks_in_batches(blocks_spend, full_node_api_1.full_node)

        # Reorg at height 5, add up to and including height 12
        new_blocks = bt.get_consecutive_blocks(
            7,
            blocks[:6],
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            seed=b"another seed",
        )

        await add_blocks_in_batches(new_blocks[-7:], full_node_api_1.full_node)

        # Spend the same coin in the new reorg chain at height 13
        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
        )

        await _validate_and_add_block(full_node_api_1.full_node.blockchain, new_blocks[-1])

        # But can't spend it twice
        new_blocks_double = bt.get_consecutive_blocks(
            1,
            new_blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
        )

        await _validate_and_add_block(
            full_node_api_1.full_node.blockchain, new_blocks_double[-1], expected_error=Err.DOUBLE_SPEND
        )

        # Now test Reorg at block 5, same spend at block height 12
        new_blocks_reorg = bt.get_consecutive_blocks(
            1,
            new_blocks[:12],
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
            seed=b"spend at 12 is ok",
        )
        await _validate_and_add_block(
            full_node_api_1.full_node.blockchain, new_blocks_reorg[-1], expected_result=AddBlockResult.ADDED_AS_ORPHAN
        )

        # Spend at height 13 is also OK (same height)
        new_blocks_reorg = bt.get_consecutive_blocks(
            1,
            new_blocks[:13],
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
            seed=b"spend at 13 is ok",
        )
        await _validate_and_add_block(
            full_node_api_1.full_node.blockchain, new_blocks_reorg[-1], expected_result=AddBlockResult.ADDED_AS_ORPHAN
        )

        # Spend at height 14 is not OK (already spend)
        new_blocks_reorg = bt.get_consecutive_blocks(
            1,
            new_blocks[:14],
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
            transaction_data=spend_bundle,
            seed=b"spend at 14 is double spend",
        )
        await _validate_and_add_block(
            full_node_api_1.full_node.blockchain,
            new_blocks_reorg[-1],
            expected_result=AddBlockResult.INVALID_BLOCK,
            expected_error=Err.DOUBLE_SPEND,
        )

    @pytest.mark.anyio
    async def test_validate_blockchain_spend_reorg_coin(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools], softfork_height: uint32
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_1_puzzlehash = WALLET_A_PUZZLE_HASHES[1]
        receiver_2_puzzlehash = WALLET_A_PUZZLE_HASHES[2]
        receiver_3_puzzlehash = WALLET_A_PUZZLE_HASHES[3]
        full_node_api_1, _, _, _, bt = two_nodes
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[2]

        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin

        assert spend_coin
        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_1_puzzlehash, spend_coin)

        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks[:5],
            seed=b"spend_reorg_coin",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
            guarantee_transaction_block=True,
        )
        await add_blocks_in_batches([new_blocks[-1]], full_node_api_1.full_node)

        coin_2 = None
        for coin in run_and_get_removals_and_additions(
            new_blocks[-1],
            test_constants.MAX_BLOCK_COST_KLVM,
            height=softfork_height,
            constants=bt.constants,
        )[1]:
            if coin.puzzle_hash == receiver_1_puzzlehash:
                coin_2 = coin
                break
        assert coin_2 is not None

        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_2_puzzlehash, coin_2)

        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks[:6],
            seed=b"spend_reorg_coin",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
            guarantee_transaction_block=True,
        )
        await add_blocks_in_batches(new_blocks, full_node_api_1.full_node)

        coin_3 = None
        for coin in run_and_get_removals_and_additions(
            new_blocks[-1],
            test_constants.MAX_BLOCK_COST_KLVM,
            height=softfork_height,
            constants=bt.constants,
        )[1]:
            if coin.puzzle_hash == receiver_2_puzzlehash:
                coin_3 = coin
                break
        assert coin_3 is not None

        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_3_puzzlehash, coin_3)

        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks[:7],
            seed=b"spend_reorg_coin",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
            guarantee_transaction_block=True,
        )
        await add_blocks_in_batches(new_blocks, full_node_api_1.full_node)

    @pytest.mark.anyio
    async def test_validate_blockchain_spend_reorg_cb_coin(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 15
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_1_puzzlehash = WALLET_A_PUZZLE_HASHES[1]
        full_node_api_1, _, _, _, bt = two_nodes
        blocks = bt.get_consecutive_blocks(num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash)
        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Spends a coinbase created in reorg
        new_blocks = bt.get_consecutive_blocks(
            5,
            blocks[:6],
            seed=b"reorg cb coin",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
        )

        await add_blocks_in_batches(new_blocks, full_node_api_1.full_node)

        spend_block = new_blocks[-1]
        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None
        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_1_puzzlehash, spend_coin)

        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks,
            seed=b"reorg cb coin",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
            guarantee_transaction_block=True,
        )
        await add_blocks_in_batches(new_blocks, full_node_api_1.full_node)

    @pytest.mark.anyio
    async def test_validate_blockchain_spend_reorg_since_genesis(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_1_puzzlehash = WALLET_A_PUZZLE_HASHES[1]
        full_node_api_1, _, _, _, bt = two_nodes
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        spend_block = blocks[-1]
        spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None
        spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_1_puzzlehash, spend_coin)

        new_blocks = bt.get_consecutive_blocks(
            1, blocks, seed=b"", farmer_reward_puzzle_hash=coinbase_puzzlehash, transaction_data=spend_bundle
        )
        await _validate_and_add_block(full_node_api_1.full_node.blockchain, new_blocks[-1])

        # Spends a coin in a genesis reorg, that was already spent
        new_blocks = bt.get_consecutive_blocks(
            12,
            [],
            seed=b"reorg since genesis",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
        )

        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks,
            seed=b"reorg since genesis",
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=spend_bundle,
        )

        await add_blocks_in_batches(new_blocks, full_node_api_1.full_node)

    @pytest.mark.anyio
    async def test_assert_my_coin_id(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent

        spend_block = blocks[2]
        bad_block = blocks[3]
        spend_coin = None
        bad_spend_coin = None
        for coin in spend_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin = coin
        assert spend_coin is not None
        for coin in bad_block.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                bad_spend_coin = coin
        assert bad_spend_coin is not None
        valid_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_MY_COIN_ID, [spend_coin.name()])
        valid_dic = {valid_cvp.opcode: [valid_cvp]}
        bad_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_MY_COIN_ID, [bad_spend_coin.name()])

        bad_dic = {bad_cvp.opcode: [bad_cvp]}
        bad_spend_bundle = wallet_a.generate_signed_transaction(uint64(1000), receiver_puzzlehash, spend_coin, bad_dic)

        valid_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin, valid_dic
        )

        assert bad_spend_bundle is not None
        assert valid_spend_bundle is not None

        # Invalid block bundle
        # Create another block that includes our transaction
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=bad_spend_bundle,
            guarantee_transaction_block=True,
        )

        # Try to validate that block
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_MY_COIN_ID_FAILED
        )

        # Valid block bundle
        # Create another block that includes our transaction
        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=valid_spend_bundle,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_coin_announcement_consumed(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        block2 = blocks[3]

        spend_coin_block_1 = None
        spend_coin_block_2 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None
        for coin in block2.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_2 = coin
        assert spend_coin_block_2 is not None

        # This condition requires block2 coinbase to be spent
        block1_cvp = ConditionWithArgs(
            ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT,
            [AssertCoinAnnouncement(asserted_id=spend_coin_block_2.name(), asserted_msg=b"test").msg_calc],
        )
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # This condition requires block1 coinbase to be spent
        block2_cvp = ConditionWithArgs(
            ConditionOpcode.CREATE_COIN_ANNOUNCEMENT,
            [b"test"],
        )
        block2_dic = {block2_cvp.opcode: [block2_cvp]}
        block2_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_2, block2_dic
        )

        # Invalid block bundle
        assert block1_spend_bundle is not None
        # Create another block that includes our transaction
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )

        # Try to validate that block
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_ANNOUNCE_CONSUMED_FAILED
        )

        # bundle_together contains both transactions
        bundle_together = SpendBundle.aggregate([block1_spend_bundle, block2_spend_bundle])

        # Create another block that includes our transaction
        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=bundle_together,
            guarantee_transaction_block=True,
        )

        # Try to validate newly created block
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_puzzle_announcement_consumed(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        block2 = blocks[3]

        spend_coin_block_1 = None
        spend_coin_block_2 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None
        for coin in block2.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_2 = coin
        assert spend_coin_block_2 is not None

        # This condition requires block2 coinbase to be spent
        block1_cvp = ConditionWithArgs(
            ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT,
            [AssertPuzzleAnnouncement(asserted_ph=spend_coin_block_2.puzzle_hash, asserted_msg=b"test").msg_calc],
        )
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # This condition requires block1 coinbase to be spent
        block2_cvp = ConditionWithArgs(
            ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT,
            [b"test"],
        )
        block2_dic = {block2_cvp.opcode: [block2_cvp]}
        block2_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_2, block2_dic
        )

        # Invalid block bundle
        assert block1_spend_bundle is not None
        # Create another block that includes our transaction
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )

        # Try to validate that block
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_ANNOUNCE_CONSUMED_FAILED
        )

        # bundle_together contains both transactions
        bundle_together = SpendBundle.aggregate([block1_spend_bundle, block2_spend_bundle])

        # Create another block that includes our transaction
        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=bundle_together,
            guarantee_transaction_block=True,
        )

        # Try to validate newly created block
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_height_absolute(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        spend_coin_block_1 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None

        # This condition requires block1 coinbase to be spent after index 10
        block1_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_HEIGHT_ABSOLUTE, [int_to_bytes(10)])
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # program that will be sent too early
        assert block1_spend_bundle is not None
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )

        # Try to validate that block at index 10
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_HEIGHT_ABSOLUTE_FAILED
        )

        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

        # At index 11, it can be spent
        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_height_relative(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 11
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        spend_coin_block_1 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None

        # This condition requires block1 coinbase to be spent after index 11
        # This condition requires block1 coinbase to be spent more than 10 block after it was farmed
        # block index has to be greater than (2 + 9 = 11)
        block1_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_HEIGHT_RELATIVE, [int_to_bytes(9)])
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # program that will be sent too early
        assert block1_spend_bundle is not None
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )

        # Try to validate that block at index 11
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_HEIGHT_RELATIVE_FAILED
        )

        new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

        # At index 12, it can be spent
        new_blocks = bt.get_consecutive_blocks(
            1,
            new_blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_seconds_relative(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        spend_coin_block_1 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None

        # This condition requires block1 coinbase to be spent 300 seconds after coin creation
        block1_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_SECONDS_RELATIVE, [int_to_bytes(300)])
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # program that will be sent to early
        assert block1_spend_bundle is not None
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            time_per_block=20,
            guarantee_transaction_block=True,
        )

        # Try to validate that block before 300 sec
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_SECONDS_RELATIVE_FAILED
        )

        # we compare the timestamp against the previous transaction block, so in
        # order to progress the timestamp, we need to farm one more block
        blocks.extend(
            bt.get_consecutive_blocks(
                1,
                blocks,
                farmer_reward_puzzle_hash=coinbase_puzzlehash,
                guarantee_transaction_block=True,
                time_per_block=301,
            )
        )
        await _validate_and_add_block(full_node_1.blockchain, blocks[-1])

        valid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
            time_per_block=301,
        )
        await _validate_and_add_block(full_node_1.blockchain, valid_new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_seconds_absolute(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        spend_coin_block_1 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None

        # This condition requires block1 coinbase to be spent after 30 seconds from now
        assert blocks[-1].foliage_transaction_block is not None
        current_time_plus3 = uint64(blocks[-1].foliage_transaction_block.timestamp + 30)
        block1_cvp = ConditionWithArgs(ConditionOpcode.ASSERT_SECONDS_ABSOLUTE, [int_to_bytes(current_time_plus3)])
        block1_dic = {block1_cvp.opcode: [block1_cvp]}
        block1_spend_bundle = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic
        )

        # program that will be sent to early
        assert block1_spend_bundle is not None
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            time_per_block=20,
            guarantee_transaction_block=True,
        )

        # Try to validate that block before 30 sec
        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.ASSERT_SECONDS_ABSOLUTE_FAILED
        )

        # we compare the timestamp against the previous transaction block, so in
        # order to progress the timestamp, we need to farm one more block
        blocks.extend(
            bt.get_consecutive_blocks(
                1,
                blocks,
                farmer_reward_puzzle_hash=coinbase_puzzlehash,
                guarantee_transaction_block=True,
                time_per_block=30,
            )
        )
        await _validate_and_add_block(full_node_1.blockchain, blocks[-1])

        valid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle,
            guarantee_transaction_block=True,
            time_per_block=31,
        )
        await _validate_and_add_block(full_node_1.blockchain, valid_new_blocks[-1])

    @pytest.mark.anyio
    async def test_assert_fee_condition(
        self, two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools]
    ) -> None:
        num_blocks = 10
        wallet_a = WALLET_A
        coinbase_puzzlehash = WALLET_A_PUZZLE_HASHES[0]
        receiver_puzzlehash = BURN_PUZZLE_HASH
        full_node_api_1, _, _, _, bt = two_nodes
        full_node_1 = full_node_api_1.full_node
        # Farm blocks
        blocks = bt.get_consecutive_blocks(
            num_blocks, farmer_reward_puzzle_hash=coinbase_puzzlehash, guarantee_transaction_block=True
        )

        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Coinbase that gets spent
        block1 = blocks[2]
        spend_coin_block_1 = None
        for coin in block1.get_included_reward_coins():
            if coin.puzzle_hash == coinbase_puzzlehash:
                spend_coin_block_1 = coin
        assert spend_coin_block_1 is not None

        # This condition requires fee to be 10 mojo
        cvp_fee = ConditionWithArgs(ConditionOpcode.RESERVE_FEE, [int_to_bytes(10)])
        # This spend bundle has 9 mojo as fee
        block1_dic_bad = {cvp_fee.opcode: [cvp_fee]}
        block1_dic_good = {cvp_fee.opcode: [cvp_fee]}
        block1_spend_bundle_bad = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic_bad, fee=9
        )
        block1_spend_bundle_good = wallet_a.generate_signed_transaction(
            uint64(1000), receiver_puzzlehash, spend_coin_block_1, block1_dic_good, fee=10
        )
        log.warning(block1_spend_bundle_good.additions())
        log.warning(f"Spend bundle fees: {estimate_fees(block1_spend_bundle_good)}")
        invalid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle_bad,
            guarantee_transaction_block=True,
        )

        await _validate_and_add_block(
            full_node_1.blockchain, invalid_new_blocks[-1], expected_error=Err.RESERVE_FEE_CONDITION_FAILED
        )

        valid_new_blocks = bt.get_consecutive_blocks(
            1,
            blocks,
            farmer_reward_puzzle_hash=coinbase_puzzlehash,
            transaction_data=block1_spend_bundle_good,
            guarantee_transaction_block=True,
        )
        await _validate_and_add_block(full_node_1.blockchain, valid_new_blocks[-1])



from __future__ import annotations

from pathlib import Path
from typing import Optional

import click
import yaml

from chik.cmds.cmd_classes import ChikCliContext
from chik.server.outbound_message import NodeType
from chik.util.config import (
    initial_config_file,
    load_defaults_for_missing_services,
    lock_and_load_config,
    save_config,
    set_peer_info,
    str2bool,
)


def configure(
    root_path: Path,
    set_farmer_peer: str,
    set_node_introducer: str,
    set_fullnode_port: str,
    set_harvester_port: str,
    set_log_level: str,
    enable_upnp: str,
    set_outbound_peer_count: str,
    set_peer_count: str,
    testnet: str,
    peer_connect_timeout: str,
    crawler_db_path: str,
    crawler_minimum_version_count: Optional[int],
    seeder_domain_name: str,
    seeder_nameserver: str,
) -> None:
    config_yaml = "config.yaml"
    with lock_and_load_config(root_path, config_yaml, fill_missing_services=True) as config:
        config.update(load_defaults_for_missing_services(config=config, config_name=config_yaml))

        change_made = False
        if set_node_introducer:
            try:
                if set_node_introducer.index(":"):
                    host, port = (
                        ":".join(set_node_introducer.split(":")[:-1]),
                        set_node_introducer.split(":")[-1],
                    )
                    config["full_node"]["introducer_peer"]["host"] = host
                    config["full_node"]["introducer_peer"]["port"] = int(port)
                    config["introducer"]["port"] = int(port)
                    print("Node introducer updated")
                    change_made = True
            except ValueError:
                print("Node introducer address must be in format [IP:Port]")
        if set_farmer_peer:
            try:
                if set_farmer_peer.index(":"):
                    host, port = (
                        ":".join(set_farmer_peer.split(":")[:-1]),
                        set_farmer_peer.split(":")[-1],
                    )
                    set_peer_info(config["harvester"], peer_type=NodeType.FARMER, peer_host=host, peer_port=int(port))
                    print("Farmer peer updated, make sure your harvester has the proper cert installed")
                    change_made = True
            except ValueError:
                print("Farmer address must be in format [IP:Port]")
        if set_fullnode_port:
            config["full_node"]["port"] = int(set_fullnode_port)
            config["full_node"]["introducer_peer"]["port"] = int(set_fullnode_port)
            set_peer_info(config["farmer"], peer_type=NodeType.FULL_NODE, peer_port=int(set_fullnode_port))
            set_peer_info(config["timelord"], peer_type=NodeType.FULL_NODE, peer_port=int(set_fullnode_port))
            set_peer_info(config["wallet"], peer_type=NodeType.FULL_NODE, peer_port=int(set_fullnode_port))
            config["wallet"]["introducer_peer"]["port"] = int(set_fullnode_port)
            config["introducer"]["port"] = int(set_fullnode_port)
            print("Default full node port updated")
            change_made = True
        if set_harvester_port:
            config["harvester"]["port"] = int(set_harvester_port)
            print("Default harvester port updated")
            change_made = True
        if set_log_level:
            levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]
            if set_log_level in levels:
                config["logging"]["log_level"] = set_log_level
                print(f"Logging level updated. Check {root_path}/log/debug.log")
                change_made = True
            else:
                print(f"Logging level not updated. Use one of: {levels}")
        if enable_upnp:
            config["full_node"]["enable_upnp"] = str2bool(enable_upnp)
            if str2bool(enable_upnp):
                print("uPnP enabled")
            else:
                print("uPnP disabled")
            change_made = True
        if set_outbound_peer_count:
            config["full_node"]["target_outbound_peer_count"] = int(set_outbound_peer_count)
            print("Target outbound peer count updated")
            change_made = True
        if set_peer_count:
            config["full_node"]["target_peer_count"] = int(set_peer_count)
            print("Target peer count updated")
            change_made = True
        if testnet:
            if testnet in {"true", "t"}:
                print("Setting Testnet")
                # check if network_overrides.constants.testnet11 exists
                if (
                    "testnet11" not in config["network_overrides"]["constants"]
                    or "testnet11" not in config["network_overrides"]["config"]
                ):
                    print("Testnet11 constants missing. Adding to config...")
                    initial_config_str: str = initial_config_file("config.yaml")
                    initial_config = yaml.safe_load(initial_config_str)
                    if "testnet11" not in config["network_overrides"]["constants"]:
                        config["network_overrides"]["constants"]["testnet11"] = initial_config["network_overrides"][
                            "constants"
                        ]["testnet11"]

                    if "testnet11" not in config["network_overrides"]["config"]:
                        config["network_overrides"]["config"]["testnet11"] = initial_config["network_overrides"][
                            "config"
                        ]["testnet11"]

                testnet_port = "59678"
                testnet_introducer = "introducer-testnet11.chiknetwork.com"
                testnet_dns_introducer = "dns-introducer-testnet11.chiknetwork.com"
                bootstrap_peers = ["testnet11-node-us-west-2.chiknetwork.com"]
                testnet = "testnet11"
                config["full_node"]["peers_file_path"] = "db/peers-testnet11.dat"
                config["wallet"]["wallet_peers_file_path"] = "wallet/db/wallet_peers-testnet11.dat"
                config["full_node"]["port"] = int(testnet_port)
                if config["full_node"]["introducer_peer"] is None:
                    config["full_node"]["introducer_peer"] = {}
                assert config["full_node"]["introducer_peer"] is not None  # mypy
                if config["wallet"]["introducer_peer"] is None:
                    config["wallet"]["introducer_peer"] = {}
                assert config["wallet"]["introducer_peer"] is not None  # mypy
                config["full_node"]["introducer_peer"]["port"] = int(testnet_port)
                set_peer_info(config["farmer"], peer_type=NodeType.FULL_NODE, peer_port=int(testnet_port))
                set_peer_info(config["timelord"], peer_type=NodeType.FULL_NODE, peer_port=int(testnet_port))
                set_peer_info(config["wallet"], peer_type=NodeType.FULL_NODE, peer_port=int(testnet_port))
                config["wallet"]["introducer_peer"]["port"] = int(testnet_port)
                config["introducer"]["port"] = int(testnet_port)
                config["full_node"]["introducer_peer"]["host"] = testnet_introducer
                config["full_node"]["dns_servers"] = [testnet_dns_introducer]
                config["wallet"]["introducer_peer"]["host"] = testnet_introducer
                config["wallet"]["dns_servers"] = [testnet_dns_introducer]
                config["selected_network"] = testnet
                config["harvester"]["selected_network"] = testnet
                config["pool"]["selected_network"] = testnet
                config["farmer"]["selected_network"] = testnet
                config["timelord"]["selected_network"] = testnet
                config["full_node"]["selected_network"] = testnet
                config["ui"]["selected_network"] = testnet
                config["introducer"]["selected_network"] = testnet
                config["wallet"]["selected_network"] = testnet
                config["data_layer"]["selected_network"] = testnet

                if "seeder" in config:
                    config["seeder"]["port"] = int(testnet_port)
                    config["seeder"]["other_peers_port"] = int(testnet_port)
                    config["seeder"]["selected_network"] = testnet
                    config["seeder"]["bootstrap_peers"] = bootstrap_peers

                print("Default full node port, introducer and network setting updated")
                change_made = True

            elif testnet in {"false", "f"}:
                print("Setting Mainnet")
                mainnet_port = "9678"
                mainnet_introducer = "introducer.chiknetwork.com"
                mainnet_dns_introducer = "dns-introducer.chiknetwork.com"
                bootstrap_peers = ["node.chiknetwork.com"]
                net = "mainnet"
                config["full_node"]["peers_file_path"] = "db/peers.dat"
                config["wallet"]["wallet_peers_file_path"] = "wallet/db/wallet_peers.dat"
                config["full_node"]["port"] = int(mainnet_port)
                config["full_node"]["introducer_peer"]["port"] = int(mainnet_port)
                set_peer_info(config["farmer"], peer_type=NodeType.FULL_NODE, peer_port=int(mainnet_port))
                set_peer_info(config["timelord"], peer_type=NodeType.FULL_NODE, peer_port=int(mainnet_port))
                set_peer_info(config["wallet"], peer_type=NodeType.FULL_NODE, peer_port=int(mainnet_port))
                config["wallet"]["introducer_peer"]["port"] = int(mainnet_port)
                config["introducer"]["port"] = int(mainnet_port)
                config["full_node"]["introducer_peer"]["host"] = mainnet_introducer
                config["full_node"]["dns_servers"] = [mainnet_dns_introducer]
                config["wallet"]["introducer_peer"]["host"] = mainnet_introducer
                config["wallet"]["dns_servers"] = [mainnet_dns_introducer]
                config["selected_network"] = net
                config["harvester"]["selected_network"] = net
                config["pool"]["selected_network"] = net
                config["farmer"]["selected_network"] = net
                config["timelord"]["selected_network"] = net
                config["full_node"]["selected_network"] = net
                config["ui"]["selected_network"] = net
                config["introducer"]["selected_network"] = net
                config["wallet"]["selected_network"] = net
                config["data_layer"]["selected_network"] = net

                if "seeder" in config:
                    config["seeder"]["port"] = int(mainnet_port)
                    config["seeder"]["other_peers_port"] = int(mainnet_port)
                    config["seeder"]["selected_network"] = net
                    config["seeder"]["bootstrap_peers"] = bootstrap_peers

                print("Default full node port, introducer and network setting updated")
                change_made = True
            else:
                print("Please choose True or False")

        if peer_connect_timeout:
            config["full_node"]["peer_connect_timeout"] = int(peer_connect_timeout)
            change_made = True

        if crawler_db_path is not None and "seeder" in config:
            config["seeder"]["crawler_db_path"] = crawler_db_path
            change_made = True

        if crawler_minimum_version_count is not None and "seeder" in config:
            config["seeder"]["minimum_version_count"] = crawler_minimum_version_count
            change_made = True

        if seeder_domain_name is not None and "seeder" in config:
            config["seeder"]["domain_name"] = seeder_domain_name
            change_made = True

        if seeder_nameserver is not None and "seeder" in config:
            config["seeder"]["nameserver"] = seeder_nameserver
            change_made = True

        if change_made:
            print("Restart any running chik services for changes to take effect")
            save_config(root_path, "config.yaml", config)


@click.command("configure", help="Modify configuration", no_args_is_help=True)
@click.option(
    "--testnet",
    "-t",
    help="configures for connection to testnet",
    type=click.Choice(["true", "t", "false", "f"]),
)
@click.option("--set-node-introducer", help="Set the introducer for node - IP:Port", type=str)
@click.option("--set-farmer-peer", help="Set the farmer peer for harvester - IP:Port", type=str)
@click.option(
    "--set-fullnode-port",
    help="Set the port to use for the fullnode, useful for testing",
    type=str,
)
@click.option(
    "--set-harvester-port",
    help="Set the port to use for the harvester, useful for testing",
    type=str,
)
@click.option(
    "--set-log-level",
    "--log-level",
    "-log-level",
    help="Set the instance log level",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
)
@click.option(
    "--enable-upnp",
    "--upnp",
    "-upnp",
    help="Enable or disable uPnP",
    type=click.Choice(["true", "t", "false", "f"]),
)
@click.option(
    "--set_outbound-peer-count",
    help="Update the target outbound peer count (default 8)",
    type=str,
)
@click.option("--set-peer-count", help="Update the target peer count (default 80)", type=str)
@click.option("--set-peer-connect-timeout", help="Update the peer connect timeout (default 30)", type=str)
@click.option(
    "--crawler-db-path",
    help="configures the path to the crawler database",
    type=str,
)
@click.option(
    "--crawler-minimum-version-count",
    help="configures how many of a particular version must be seen to be reported in logs",
    type=int,
)
@click.option(
    "--seeder-domain-name",
    help="configures the seeder domain_name setting. Ex: `seeder.example.com.`",
    type=str,
)
@click.option(
    "--seeder-nameserver",
    help="configures the seeder nameserver setting. Ex: `example.com.`",
    type=str,
)
@click.pass_context
def configure_cmd(
    ctx: click.Context,
    set_farmer_peer: str,
    set_node_introducer: str,
    set_fullnode_port: str,
    set_harvester_port: str,
    set_log_level: str,
    enable_upnp: str,
    set_outbound_peer_count: str,
    set_peer_count: str,
    testnet: str,
    set_peer_connect_timeout: str,
    crawler_db_path: str,
    crawler_minimum_version_count: int,
    seeder_domain_name: str,
    seeder_nameserver: str,
) -> None:
    configure(
        ChikCliContext.set_default(ctx).root_path,
        set_farmer_peer,
        set_node_introducer,
        set_fullnode_port,
        set_harvester_port,
        set_log_level,
        enable_upnp,
        set_outbound_peer_count,
        set_peer_count,
        testnet,
        set_peer_connect_timeout,
        crawler_db_path,
        crawler_minimum_version_count,
        seeder_domain_name,
        seeder_nameserver,
    )



from __future__ import annotations

from chik._tests.util.get_name_puzzle_conditions import get_name_puzzle_conditions
from chik.consensus.cost_calculator import NPCResult
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.full_node.bundle_tools import simple_solution_generator
from chik.types.blockchain_format.program import INFINITE_COST
from chik.types.generator_types import BlockGenerator
from chik.types.spend_bundle import SpendBundle
from chik.util.ints import uint64


def cost_of_spend_bundle(spend_bundle: SpendBundle) -> int:
    program: BlockGenerator = simple_solution_generator(spend_bundle)
    # always use the post soft-fork2 semantics
    npc_result: NPCResult = get_name_puzzle_conditions(
        program,
        INFINITE_COST,
        mempool_mode=True,
        height=DEFAULT_CONSTANTS.HARD_FORK_HEIGHT,
        constants=DEFAULT_CONSTANTS,
    )
    return uint64(0 if npc_result.conds is None else npc_result.conds.cost)



from __future__ import annotations

import contextlib
from collections.abc import AsyncIterator
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Optional, cast, final

from chik.types.blockchain_format.coin import Coin
from chik.util.action_scope import ActionScope
from chik.util.streamable import Streamable, streamable
from chik.wallet.signer_protocol import SigningResponse
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.util.tx_config import TXConfig
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

if TYPE_CHECKING:
    # Avoid a circular import here
    from chik.wallet.wallet_state_manager import WalletStateManager


@streamable
@dataclass(frozen=True)
class _StreamableWalletSideEffects(Streamable):
    transactions: list[TransactionRecord]
    signing_responses: list[SigningResponse]
    extra_spends: list[WalletSpendBundle]
    selected_coins: list[Coin]


@dataclass
class WalletSideEffects:
    transactions: list[TransactionRecord] = field(default_factory=list)
    signing_responses: list[SigningResponse] = field(default_factory=list)
    extra_spends: list[WalletSpendBundle] = field(default_factory=list)
    selected_coins: list[Coin] = field(default_factory=list)

    def __bytes__(self) -> bytes:
        return bytes(_StreamableWalletSideEffects(**self.__dict__))

    @classmethod
    def from_bytes(cls, blob: bytes) -> WalletSideEffects:
        return cls(**_StreamableWalletSideEffects.from_bytes(blob).__dict__)


@final
@dataclass(frozen=True)
class WalletActionConfig:
    push: bool
    merge_spends: bool
    sign: Optional[bool]
    additional_signing_responses: list[SigningResponse]
    extra_spends: list[WalletSpendBundle]
    tx_config: TXConfig

    def adjust_for_side_effects(self, side_effects: WalletSideEffects) -> WalletActionConfig:
        return replace(
            self,
            tx_config=replace(
                self.tx_config,
                excluded_coin_ids=[*self.tx_config.excluded_coin_ids, *(c.name() for c in side_effects.selected_coins)],
            ),
        )


WalletActionScope = ActionScope[WalletSideEffects, WalletActionConfig]


@contextlib.asynccontextmanager
async def new_wallet_action_scope(
    wallet_state_manager: WalletStateManager,
    tx_config: TXConfig,
    push: bool = False,
    merge_spends: bool = True,
    sign: Optional[bool] = None,
    additional_signing_responses: list[SigningResponse] = [],
    extra_spends: list[WalletSpendBundle] = [],
) -> AsyncIterator[WalletActionScope]:
    async with ActionScope.new_scope(
        WalletSideEffects,
        WalletActionConfig(push, merge_spends, sign, additional_signing_responses, extra_spends, tx_config),
    ) as self:
        self = cast(WalletActionScope, self)
        async with self.use() as interface:
            interface.side_effects.signing_responses = additional_signing_responses.copy()
            interface.side_effects.extra_spends = extra_spends.copy()

        yield self

    self.side_effects.transactions = await wallet_state_manager.add_pending_transactions(
        self.side_effects.transactions,
        push=push,
        merge_spends=merge_spends,
        sign=sign,
        additional_signing_responses=self.side_effects.signing_responses,
        extra_spends=self.side_effects.extra_spends,
    )



from __future__ import annotations

from typing import Any

from chik.consensus.default_constants import update_testnet_overrides


def test_testnet11() -> None:
    overrides: dict[str, Any] = {}
    update_testnet_overrides("testnet11", overrides)
    assert overrides == {
        "SOFT_FORK6_HEIGHT": 2000000,
    }


def test_mainnet() -> None:
    overrides: dict[str, Any] = {}
    update_testnet_overrides("mainnet", overrides)
    assert overrides == {}



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional, Union

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_record import CoinRecord
from chik.util.ints import uint8, uint32, uint64
from chik.util.streamable import VersionedBlob
from chik.wallet.puzzles.clawback.metadata import ClawbackMetadata, ClawbackVersion
from chik.wallet.util.wallet_types import CoinType, StreamableWalletIdentifier, WalletType
from chik.wallet.vc_wallet.cr_cat_drivers import CRCATMetadata, CRCATVersion

MetadataTypes = Union[ClawbackMetadata, CRCATMetadata]


@dataclass(frozen=True)
class WalletCoinRecord:
    """
    These are values that correspond to a CoinName that are used
    in keeping track of the unspent database.
    """

    coin: Coin
    confirmed_block_height: uint32
    spent_block_height: uint32
    spent: bool
    coinbase: bool
    wallet_type: WalletType
    wallet_id: int
    # Cannot include new attributes in the hash since they will change the coin order in a set.
    # The launcher coin ID will change and will break all hardcode offer tests in CAT/NFT/DL, etc.
    # TODO Change hardcode offer in unit tests
    coin_type: CoinType = field(default=CoinType.NORMAL, hash=False)
    metadata: Optional[VersionedBlob] = field(default=None, hash=False)

    def wallet_identifier(self) -> StreamableWalletIdentifier:
        return StreamableWalletIdentifier(uint32(self.wallet_id), uint8(self.wallet_type))

    def parsed_metadata(self) -> MetadataTypes:
        if self.metadata is None:
            raise ValueError("Can't parse None metadata")
        if self.coin_type == CoinType.CLAWBACK and self.metadata.version == ClawbackVersion.V1.value:
            return ClawbackMetadata.from_bytes(self.metadata.blob)
        if (
            self.coin_type in {CoinType.CRCAT_PENDING, CoinType.CRCAT}
            and self.metadata.version == CRCATVersion.V1.value
        ):
            return CRCATMetadata.from_bytes(self.metadata.blob)
        raise ValueError(f"Unknown metadata {self.metadata} for coin_type {self.coin_type}")

    def name(self) -> bytes32:
        return self.coin.name()

    def to_coin_record(self, timestamp: uint64) -> CoinRecord:
        return CoinRecord(self.coin, self.confirmed_block_height, self.spent_block_height, self.coinbase, timestamp)

    def to_json_dict_parsed_metadata(self) -> dict[str, Any]:
        # TODO: Merge wallet_type and wallet_id into `wallet_identifier`, make `spent` an attribute based
        #  on `spent_height` make `WalletCoinRecord` streamable and use Streamable.to_json_dict as base here if we have
        #  streamable enums.
        return {
            **self.coin.to_json_dict(),
            "id": "0x" + self.name().hex(),
            "type": int(self.coin_type),
            "wallet_identifier": self.wallet_identifier().to_json_dict(),
            "metadata": None if self.metadata is None else self.parsed_metadata().to_json_dict(),
            "confirmed_height": self.confirmed_block_height,
            "spent_height": self.spent_block_height,
            "coinbase": self.coinbase,
        }



# All of these rate limits scale with the number of transactions so the aggregate amounts are higher
from __future__ import annotations

import copy
import dataclasses
import functools
from typing import Any, Optional

from chik.protocols.protocol_message_types import ProtocolMessageTypes
from chik.protocols.shared_protocol import Capability

compose_rate_limits_cache: dict[int, dict[str, Any]] = {}


# this class is used to configure the *rate* limit for a message type. The
# limits are counts and size per 60 seconds.
@dataclasses.dataclass(frozen=True)
class RLSettings:
    frequency: int  # Max request per time period (ie 1 min)
    max_size: int  # Max size of each request
    max_total_size: Optional[int] = None  # Max cumulative size of all requests in that period


# this class is used to indicate that a message type is not subject to a rate
# limit, but just a per-message size limit. This may be appropriate for response
# messages that are implicitly limited by their corresponding request message
@dataclasses.dataclass(frozen=True)
class Unlimited:
    max_size: int  # Max size of each request


def get_rate_limits_to_use(our_capabilities: list[Capability], peer_capabilities: list[Capability]) -> dict[str, Any]:
    # This will use the newest possible rate limits that both peers support. At this time there are only two
    # options, v1 and v2.

    if Capability.RATE_LIMITS_V2 in our_capabilities and Capability.RATE_LIMITS_V2 in peer_capabilities:
        # Use V2 rate limits
        if 2 in compose_rate_limits_cache:
            return compose_rate_limits_cache[2]
        composed = compose_rate_limits(rate_limits[1], rate_limits[2])
        compose_rate_limits_cache[2] = composed
        return composed
    else:
        # Use V1 rate limits
        return rate_limits[1]


def compose_rate_limits(old_rate_limits: dict[str, Any], new_rate_limits: dict[str, Any]) -> dict[str, Any]:
    # Composes two rate limits dicts, so that the newer values override the older values
    final_rate_limits: dict[str, Any] = copy.deepcopy(new_rate_limits)
    categories: list[str] = ["rate_limits_tx", "rate_limits_other"]
    all_new_msgs_lists: list[list[ProtocolMessageTypes]] = [
        list(new_rate_limits[category].keys()) for category in categories
    ]
    all_new_msgs: list[ProtocolMessageTypes] = functools.reduce(lambda a, b: a + b, all_new_msgs_lists)
    for old_cat, mapping in old_rate_limits.items():
        if old_cat in categories:
            for old_protocol_msg, old_rate_limit_value in mapping.items():
                if old_protocol_msg not in all_new_msgs:
                    if old_cat not in final_rate_limits:
                        final_rate_limits[old_cat] = {}
                    final_rate_limits[old_cat][old_protocol_msg] = old_rate_limit_value
    return final_rate_limits


# Each number in this dict corresponds to a specific version of rate limits (1, 2,  etc).
# Version 1 includes the original limits for chik software from versions 1.0 to 1.4.
rate_limits = {
    1: {
        "default_settings": RLSettings(100, 1024 * 1024, 100 * 1024 * 1024),
        "non_tx_freq": 1000,  # There is also a freq limit for many requests
        "non_tx_max_total_size": 100 * 1024 * 1024,  # There is also a size limit for many requests
        # All transaction related apis also have an aggregate limit
        "rate_limits_tx": {
            ProtocolMessageTypes.new_transaction: RLSettings(5000, 100, 5000 * 100),
            ProtocolMessageTypes.request_transaction: RLSettings(5000, 100, 5000 * 100),
            ProtocolMessageTypes.respond_transaction: RLSettings(
                5000, 1 * 1024 * 1024, 20 * 1024 * 1024
            ),  # TODO: check this
            ProtocolMessageTypes.send_transaction: RLSettings(5000, 1024 * 1024),
            ProtocolMessageTypes.transaction_ack: RLSettings(5000, 2048),
        },
        # All non-transaction apis also have an aggregate limit
        "rate_limits_other": {
            ProtocolMessageTypes.handshake: RLSettings(5, 10 * 1024, 5 * 10 * 1024),
            ProtocolMessageTypes.harvester_handshake: RLSettings(5, 1024 * 1024),
            ProtocolMessageTypes.new_signage_point_harvester: RLSettings(100, 4886),  # Size with 100 pool list
            ProtocolMessageTypes.new_proof_of_space: RLSettings(100, 2048),
            ProtocolMessageTypes.request_signatures: RLSettings(100, 2048),
            ProtocolMessageTypes.respond_signatures: RLSettings(100, 2048),
            ProtocolMessageTypes.new_signage_point: RLSettings(200, 2048),
            ProtocolMessageTypes.declare_proof_of_space: RLSettings(100, 10 * 1024),
            ProtocolMessageTypes.request_signed_values: RLSettings(100, 10 * 1024),
            ProtocolMessageTypes.farming_info: RLSettings(100, 1024),
            ProtocolMessageTypes.signed_values: RLSettings(100, 1024),
            ProtocolMessageTypes.new_peak_timelord: RLSettings(100, 20 * 1024),
            ProtocolMessageTypes.new_unfinished_block_timelord: RLSettings(100, 10 * 1024),
            ProtocolMessageTypes.new_signage_point_vdf: RLSettings(100, 100 * 1024),
            ProtocolMessageTypes.new_infusion_point_vdf: RLSettings(100, 100 * 1024),
            ProtocolMessageTypes.new_end_of_sub_slot_vdf: RLSettings(100, 100 * 1024),
            ProtocolMessageTypes.request_compact_proof_of_time: RLSettings(100, 10 * 1024),
            ProtocolMessageTypes.respond_compact_proof_of_time: RLSettings(100, 100 * 1024),
            ProtocolMessageTypes.new_peak: RLSettings(200, 512),
            ProtocolMessageTypes.request_proof_of_weight: RLSettings(5, 100),
            ProtocolMessageTypes.respond_proof_of_weight: RLSettings(5, 50 * 1024 * 1024, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_block: RLSettings(200, 100),
            ProtocolMessageTypes.reject_block: Unlimited(100),
            ProtocolMessageTypes.request_blocks: RLSettings(500, 100),
            ProtocolMessageTypes.respond_blocks: Unlimited(50 * 1024 * 1024),
            ProtocolMessageTypes.reject_blocks: Unlimited(100),
            ProtocolMessageTypes.respond_block: Unlimited(2 * 1024 * 1024),
            ProtocolMessageTypes.new_unfinished_block: RLSettings(200, 100),
            ProtocolMessageTypes.request_unfinished_block: RLSettings(200, 100),
            ProtocolMessageTypes.new_unfinished_block2: RLSettings(200, 100),
            ProtocolMessageTypes.request_unfinished_block2: RLSettings(200, 100),
            ProtocolMessageTypes.respond_unfinished_block: RLSettings(200, 2 * 1024 * 1024, 10 * 2 * 1024 * 1024),
            ProtocolMessageTypes.new_signage_point_or_end_of_sub_slot: RLSettings(200, 200),
            ProtocolMessageTypes.request_signage_point_or_end_of_sub_slot: RLSettings(200, 200),
            ProtocolMessageTypes.respond_signage_point: RLSettings(200, 50 * 1024),
            ProtocolMessageTypes.respond_end_of_sub_slot: RLSettings(100, 50 * 1024),
            ProtocolMessageTypes.request_mempool_transactions: RLSettings(5, 1024 * 1024),
            ProtocolMessageTypes.request_compact_vdf: RLSettings(200, 1024),
            ProtocolMessageTypes.respond_compact_vdf: RLSettings(200, 100 * 1024),
            ProtocolMessageTypes.new_compact_vdf: RLSettings(100, 1024),
            ProtocolMessageTypes.request_peers: RLSettings(10, 100),
            ProtocolMessageTypes.respond_peers: RLSettings(10, 1 * 1024 * 1024),
            ProtocolMessageTypes.request_puzzle_solution: RLSettings(1000, 100),
            ProtocolMessageTypes.respond_puzzle_solution: RLSettings(1000, 1024 * 1024),
            ProtocolMessageTypes.reject_puzzle_solution: RLSettings(1000, 100),
            ProtocolMessageTypes.new_peak_wallet: RLSettings(200, 300),
            ProtocolMessageTypes.request_block_header: RLSettings(500, 100),
            ProtocolMessageTypes.respond_block_header: RLSettings(500, 500 * 1024),
            ProtocolMessageTypes.reject_header_request: RLSettings(500, 100),
            ProtocolMessageTypes.request_removals: RLSettings(500, 50 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.respond_removals: RLSettings(500, 1024 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.reject_removals_request: RLSettings(500, 100),
            ProtocolMessageTypes.request_additions: RLSettings(500, 1024 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.respond_additions: RLSettings(500, 1024 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.reject_additions_request: RLSettings(500, 100),
            ProtocolMessageTypes.request_header_blocks: RLSettings(500, 100),
            ProtocolMessageTypes.reject_header_blocks: RLSettings(100, 100),
            ProtocolMessageTypes.respond_header_blocks: RLSettings(500, 2 * 1024 * 1024, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_peers_introducer: RLSettings(100, 100),
            ProtocolMessageTypes.respond_peers_introducer: RLSettings(100, 1024 * 1024),
            ProtocolMessageTypes.farm_new_block: RLSettings(200, 200),
            ProtocolMessageTypes.request_plots: RLSettings(10, 10 * 1024 * 1024),
            ProtocolMessageTypes.respond_plots: RLSettings(10, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_start: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_loaded: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_removed: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_invalid: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_keys_missing: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_duplicates: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_done: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.plot_sync_response: RLSettings(3000, 100 * 1024 * 1024),
            ProtocolMessageTypes.coin_state_update: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.register_for_ph_updates: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_to_ph_updates: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.register_for_coin_updates: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_to_coin_updates: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_remove_puzzle_subscriptions: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_remove_puzzle_subscriptions: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_remove_coin_subscriptions: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_remove_coin_subscriptions: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_puzzle_state: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_puzzle_state: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.reject_puzzle_state: RLSettings(200, 100),
            ProtocolMessageTypes.request_coin_state: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_coin_state: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.reject_coin_state: RLSettings(200, 100),
            ProtocolMessageTypes.mempool_items_added: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.mempool_items_removed: RLSettings(1000, 100 * 1024 * 1024),
            ProtocolMessageTypes.request_cost_info: RLSettings(1000, 100),
            ProtocolMessageTypes.respond_cost_info: RLSettings(1000, 1024),
            ProtocolMessageTypes.request_ses_hashes: RLSettings(2000, 1 * 1024 * 1024),
            ProtocolMessageTypes.respond_ses_hashes: RLSettings(2000, 1 * 1024 * 1024),
            ProtocolMessageTypes.request_children: RLSettings(2000, 1024 * 1024),
            ProtocolMessageTypes.respond_children: RLSettings(2000, 1 * 1024 * 1024),
        },
    },
    2: {
        "default_settings": RLSettings(100, 1024 * 1024, 100 * 1024 * 1024),
        "non_tx_freq": 1000,  # There is also a freq limit for many requests
        "non_tx_max_total_size": 100 * 1024 * 1024,  # There is also a size limit for many requests
        "rate_limits_tx": {
            ProtocolMessageTypes.request_block_header: RLSettings(500, 100),
            ProtocolMessageTypes.respond_block_header: RLSettings(500, 500 * 1024),
            ProtocolMessageTypes.reject_header_request: RLSettings(500, 100),
            ProtocolMessageTypes.request_removals: RLSettings(5000, 50 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.respond_removals: RLSettings(5000, 1024 * 1024, 10 * 1024 * 1024),
            ProtocolMessageTypes.reject_removals_request: RLSettings(500, 100),
            ProtocolMessageTypes.request_additions: RLSettings(50000, 100 * 1024 * 1024),
            ProtocolMessageTypes.respond_additions: RLSettings(50000, 100 * 1024 * 1024),
            ProtocolMessageTypes.reject_additions_request: RLSettings(500, 100),
            ProtocolMessageTypes.reject_header_blocks: RLSettings(1000, 100),
            ProtocolMessageTypes.respond_header_blocks: RLSettings(5000, 2 * 1024 * 1024),
            ProtocolMessageTypes.request_block_headers: RLSettings(5000, 100),
            ProtocolMessageTypes.reject_block_headers: RLSettings(1000, 100),
            ProtocolMessageTypes.respond_block_headers: RLSettings(5000, 2 * 1024 * 1024),
            ProtocolMessageTypes.request_ses_hashes: RLSettings(2000, 1 * 1024 * 1024),
            ProtocolMessageTypes.respond_ses_hashes: RLSettings(2000, 1 * 1024 * 1024),
            ProtocolMessageTypes.request_children: RLSettings(2000, 1024 * 1024),
            ProtocolMessageTypes.respond_children: RLSettings(2000, 1 * 1024 * 1024),
            ProtocolMessageTypes.request_puzzle_solution: RLSettings(5000, 100),
            ProtocolMessageTypes.respond_puzzle_solution: RLSettings(5000, 1024 * 1024),
            ProtocolMessageTypes.reject_puzzle_solution: RLSettings(5000, 100),
            ProtocolMessageTypes.none_response: RLSettings(500, 100),
            ProtocolMessageTypes.error: RLSettings(50000, 100),
        },
        "rate_limits_other": {  # These will have a lower cap since they don't scale with high TPS (NON_TX_FREQ)
            ProtocolMessageTypes.request_header_blocks: RLSettings(5000, 100),
        },
    },
}



from __future__ import annotations

import logging
import time
import traceback
from typing import TYPE_CHECKING, Any, Optional

from chik_rs import G1Element, G2Element
from typing_extensions import Unpack

from chik.server.ws_connection import WSChikConnection
from chik.types.blockchain_format.coin import Coin, coin_as_list
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend
from chik.util.byte_types import hexstr_to_bytes
from chik.util.hash import std_hash
from chik.util.ints import uint8, uint32, uint64, uint128
from chik.util.streamable import VersionedBlob
from chik.wallet.cat_wallet.cat_info import CATCoinData, CRCATInfo
from chik.wallet.cat_wallet.cat_utils import CAT_MOD_HASH, CAT_MOD_HASH_HASH, construct_cat_puzzle
from chik.wallet.cat_wallet.cat_wallet import CATWallet
from chik.wallet.coin_selection import select_coins
from chik.wallet.conditions import (
    Condition,
    ConditionValidTimes,
    CreateCoinAnnouncement,
    CreatePuzzleAnnouncement,
    UnknownCondition,
    parse_timelock_info,
)
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.outer_puzzles import AssetType
from chik.wallet.payment import Payment
from chik.wallet.puzzle_drivers import PuzzleInfo
from chik.wallet.trading.offer import Offer
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.uncurried_puzzle import uncurry_puzzle
from chik.wallet.util.compute_hints import compute_spend_hints_and_additions
from chik.wallet.util.compute_memos import compute_memos
from chik.wallet.util.query_filter import HashFilter
from chik.wallet.util.transaction_type import TransactionType
from chik.wallet.util.wallet_sync_utils import fetch_coin_spend_for_coin_state
from chik.wallet.util.wallet_types import CoinType, WalletType
from chik.wallet.vc_wallet.cr_cat_drivers import (
    CRCAT,
    CRCATMetadata,
    CRCATVersion,
    ProofsChecker,
    construct_cr_layer_hash,
    construct_pending_approval_state,
)
from chik.wallet.vc_wallet.vc_drivers import VerifiedCredential
from chik.wallet.vc_wallet.vc_wallet import VCWallet
from chik.wallet.wallet import Wallet
from chik.wallet.wallet_action_scope import WalletActionScope
from chik.wallet.wallet_coin_record import MetadataTypes, WalletCoinRecord
from chik.wallet.wallet_info import WalletInfo
from chik.wallet.wallet_protocol import GSTOptionalArgs, WalletProtocol
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

if TYPE_CHECKING:
    from chik.wallet.wallet_state_manager import WalletStateManager


class CRCATWallet(CATWallet):
    wallet_state_manager: WalletStateManager
    log: logging.Logger
    wallet_info: WalletInfo
    info: CRCATInfo
    standard_wallet: Wallet

    @staticmethod
    def default_wallet_name_for_unknown_cat(limitations_program_hash_hex: str) -> str:
        return f"CAT {limitations_program_hash_hex[:16]}..."

    @property
    def cost_of_single_tx(self) -> int:
        return 78000000  # Estimate measured in testing

    @staticmethod
    async def create_new_cat_wallet(
        wallet_state_manager: WalletStateManager,
        wallet: Wallet,
        cat_tail_info: dict[str, Any],
        amount: uint64,
        action_scope: WalletActionScope,
        fee: uint64 = uint64(0),
        name: Optional[str] = None,
        push: bool = False,
    ) -> CATWallet:  # pragma: no cover
        raise NotImplementedError("create_new_cat_wallet is a legacy method and is not available on CR-CAT wallets")

    @staticmethod
    async def get_or_create_wallet_for_cat(
        wallet_state_manager: WalletStateManager,
        wallet: Wallet,
        limitations_program_hash_hex: str,
        name: Optional[str] = None,
        authorized_providers: Optional[list[bytes32]] = None,
        proofs_checker: Optional[ProofsChecker] = None,
    ) -> CRCATWallet:
        if authorized_providers is None or proofs_checker is None:  # pragma: no cover
            raise ValueError("get_or_create_wallet_for_cat was call on CRCATWallet without proper arguments")
        self = CRCATWallet()
        self.standard_wallet = wallet
        if name is None:
            name = self.default_wallet_name_for_unknown_cat(limitations_program_hash_hex)
        self.log = logging.getLogger(name)

        tail_hash = bytes32.from_hexstr(limitations_program_hash_hex)

        for id, w in wallet_state_manager.wallets.items():
            if w.type() == CRCATWallet.type():
                assert isinstance(w, CRCATWallet)
                if w.get_asset_id() == limitations_program_hash_hex:
                    self.log.warning("Not creating wallet for already existing CR-CAT wallet")
                    return w

        self.wallet_state_manager = wallet_state_manager

        self.info = CRCATInfo(tail_hash, None, authorized_providers, proofs_checker)
        info_as_string = bytes(self.info).hex()
        self.wallet_info = await wallet_state_manager.user_store.create_wallet(name, WalletType.CRCAT, info_as_string)

        await self.wallet_state_manager.add_new_wallet(self)
        return self

    @classmethod
    async def create_from_puzzle_info(
        cls,
        wallet_state_manager: WalletStateManager,
        wallet: Wallet,
        puzzle_driver: PuzzleInfo,
        name: Optional[str] = None,
        # We're hinting this as Any for mypy by should explore adding this to the wallet protocol and hinting properly
        potential_subclasses: dict[AssetType, Any] = {},
    ) -> Any:
        cr_layer: Optional[PuzzleInfo] = puzzle_driver.also()
        if cr_layer is None:  # pragma: no cover
            raise ValueError("create_from_puzzle_info called on CRCATWallet with a non CR-CAT puzzle driver")
        return await cls.get_or_create_wallet_for_cat(
            wallet_state_manager,
            wallet,
            puzzle_driver["tail"].hex(),
            name,
            [bytes32(provider) for provider in cr_layer["authorized_providers"]],
            ProofsChecker.from_program(uncurry_puzzle(cr_layer["proofs_checker"])),
        )

    @staticmethod
    async def create(
        wallet_state_manager: WalletStateManager,
        wallet: Wallet,
        wallet_info: WalletInfo,
    ) -> CRCATWallet:
        self = CRCATWallet()

        self.log = logging.getLogger(__name__)
        self.wallet_state_manager = wallet_state_manager
        self.wallet_info = wallet_info
        self.standard_wallet = wallet
        self.info = CRCATInfo.from_bytes(hexstr_to_bytes(self.wallet_info.data))
        return self

    @classmethod
    async def convert_to_cr(
        cls,
        cat_wallet: CATWallet,
        authorized_providers: list[bytes32],
        proofs_checker: ProofsChecker,
    ) -> None:
        replace_self = cls()
        replace_self.standard_wallet = cat_wallet.standard_wallet
        replace_self.log = logging.getLogger(cat_wallet.get_name())
        replace_self.log.info(f"Converting CAT wallet {cat_wallet.id()} to CR-CAT wallet")
        replace_self.wallet_state_manager = cat_wallet.wallet_state_manager
        replace_self.info = CRCATInfo(
            cat_wallet.cat_info.limitations_program_hash, None, authorized_providers, proofs_checker
        )
        await cat_wallet.wallet_state_manager.user_store.update_wallet(
            WalletInfo(
                cat_wallet.id(), cat_wallet.get_name(), uint8(WalletType.CRCAT.value), bytes(replace_self.info).hex()
            )
        )
        updated_wallet_info = await cat_wallet.wallet_state_manager.user_store.get_wallet_by_id(cat_wallet.id())
        assert updated_wallet_info is not None
        replace_self.wallet_info = updated_wallet_info

        cat_wallet.wallet_state_manager.wallets[cat_wallet.id()] = replace_self

    @classmethod
    def type(cls) -> WalletType:
        return WalletType.CRCAT

    def id(self) -> uint32:
        return self.wallet_info.id

    def get_asset_id(self) -> str:
        return self.info.limitations_program_hash.hex()

    async def set_tail_program(self, tail_program: str) -> None:  # pragma: no cover
        raise NotImplementedError("set_tail_program is a legacy method and is not available on CR-CAT wallets")

    async def coin_added(
        self, coin: Coin, height: uint32, peer: WSChikConnection, coin_data: Optional[CATCoinData]
    ) -> None:
        """Notification from wallet state manager that wallet has been received."""
        self.log.info(f"CR-CAT wallet has been notified that {coin.name().hex()} was added")
        try:
            coin_state = await self.wallet_state_manager.wallet_node.get_coin_state([coin.parent_coin_info], peer=peer)
            coin_spend = await fetch_coin_spend_for_coin_state(coin_state[0], peer)
            await self.add_crcat_coin(coin_spend, coin, height)
        except Exception as e:
            self.log.debug(f"Exception: {e}, traceback: {traceback.format_exc()}")

    async def add_crcat_coin(self, coin_spend: CoinSpend, coin: Coin, height: uint32) -> None:
        try:
            new_cr_cats: list[CRCAT] = CRCAT.get_next_from_coin_spend(coin_spend)
            hint_dict = {
                id: hc.hint
                for id, hc in compute_spend_hints_and_additions(coin_spend)[0].items()
                if hc.hint is not None
            }
            cr_cat: CRCAT = next(filter(lambda c: c.coin.name() == coin.name(), new_cr_cats))
            if (
                await self.wallet_state_manager.puzzle_store.get_derivation_record_for_puzzle_hash(
                    cr_cat.inner_puzzle_hash
                )
                is not None
            ):
                self.log.info(f"Found CRCAT coin {coin.name().hex()}")
                is_pending = False
            elif (
                cr_cat.inner_puzzle_hash
                == construct_pending_approval_state(
                    hint_dict[coin.name()],
                    uint64(coin.amount),
                ).get_tree_hash()
            ):
                self.log.info(f"Found pending approval CRCAT coin {coin.name().hex()}")
                is_pending = True
                created_timestamp = await self.wallet_state_manager.wallet_node.get_timestamp_for_height(uint32(height))
                spend_bundle = WalletSpendBundle([coin_spend], G2Element())
                memos = compute_memos(spend_bundle)
                # This will override the tx created in the wallet state manager
                tx_record = TransactionRecord(
                    confirmed_at_height=height,
                    created_at_time=uint64(created_timestamp),
                    to_puzzle_hash=hint_dict[coin.name()],
                    amount=uint64(coin.amount),
                    fee_amount=uint64(0),
                    confirmed=True,
                    sent=uint32(0),
                    spend_bundle=None,
                    additions=[coin],
                    removals=[coin_spend.coin],
                    wallet_id=self.id(),
                    sent_to=[],
                    trade_id=None,
                    type=uint32(TransactionType.INCOMING_CRCAT_PENDING),
                    name=coin.name(),
                    memos=list(memos.items()),
                    valid_times=ConditionValidTimes(),
                )
                await self.wallet_state_manager.tx_store.add_transaction_record(tx_record)
            else:  # pragma: no cover
                self.log.error(f"Unknown CRCAT inner puzzle, coin ID: {coin.name().hex()}")
                return None
            coin_record = WalletCoinRecord(
                coin,
                uint32(height),
                uint32(0),
                False,
                False,
                WalletType.CRCAT,
                self.id(),
                CoinType.CRCAT_PENDING if is_pending else CoinType.CRCAT,
                VersionedBlob(
                    CRCATVersion.V1.value,
                    bytes(
                        CRCATMetadata(
                            cr_cat.lineage_proof, hint_dict[coin.name()] if is_pending else cr_cat.inner_puzzle_hash
                        )
                    ),
                ),
            )
            await self.wallet_state_manager.coin_store.add_coin_record(coin_record)
        except Exception:
            # The parent is not a CAT which means we need to scrub all of its children from our DB
            self.log.error(f"Cannot add CRCAT coin: {traceback.format_exc()}")
            child_coin_records = await self.wallet_state_manager.coin_store.get_coin_records_by_parent_id(
                coin_spend.coin.name()
            )
            if len(child_coin_records) > 0:
                for record in child_coin_records:
                    if record.wallet_id == self.id():  # pragma: no cover
                        await self.wallet_state_manager.coin_store.delete_coin_record(record.coin.name())
                        # We also need to make sure there's no record of the transaction
                        await self.wallet_state_manager.tx_store.delete_transaction_record(record.coin.name())

    def require_derivation_paths(self) -> bool:
        return False

    def puzzle_for_pk(self, pubkey: G1Element) -> Program:  # pragma: no cover
        raise NotImplementedError("puzzle_for_pk is a legacy method and is not available on CR-CAT wallets")

    def puzzle_hash_for_pk(self, pubkey: G1Element) -> bytes32:  # pragma: no cover
        raise NotImplementedError("puzzle_hash_for_pk is a legacy method and is not available on CR-CAT wallets")

    async def get_new_cat_puzzle_hash(self) -> bytes32:  # pragma: no cover
        raise NotImplementedError("get_new_cat_puzzle_hash is a legacy method and is not available on CR-CAT wallets")

    async def inner_puzzle_for_cat_puzhash(self, cat_hash: bytes32) -> Program:  # pragma: no cover
        raise NotImplementedError(
            "inner_puzzle_for_cat_puzhash is a legacy method and is not available on CR-CAT wallets"
        )

    async def get_cat_spendable_coins(self, records: Optional[set[WalletCoinRecord]] = None) -> list[WalletCoinRecord]:
        result: list[WalletCoinRecord] = []

        record_list: set[WalletCoinRecord] = await self.wallet_state_manager.get_spendable_coins_for_wallet(
            self.id(), records
        )

        for record in record_list:
            crcat: CRCAT = self.coin_record_to_crcat(record)
            if crcat.lineage_proof is not None and not crcat.lineage_proof.is_none():
                result.append(record)

        return result

    async def get_confirmed_balance(self, record_list: Optional[set[WalletCoinRecord]] = None) -> uint128:
        if record_list is None:
            record_list = await self.wallet_state_manager.coin_store.get_unspent_coins_for_wallet(
                self.id(), CoinType.CRCAT
            )
        amount: uint128 = uint128(0)
        for record in record_list:
            crcat: CRCAT = self.coin_record_to_crcat(record)
            if crcat.lineage_proof is not None and not crcat.lineage_proof.is_none():
                amount = uint128(amount + record.coin.amount)

        self.log.info(f"Confirmed balance for cat wallet {self.id()} is {amount}")
        return uint128(amount)

    async def get_pending_approval_balance(self, record_list: Optional[set[WalletCoinRecord]] = None) -> uint128:
        if record_list is None:
            record_list = await self.wallet_state_manager.coin_store.get_unspent_coins_for_wallet(
                self.id(), CoinType.CRCAT_PENDING
            )
        amount: uint128 = uint128(0)
        for record in record_list:
            crcat: CRCAT = self.coin_record_to_crcat(record)
            if crcat.lineage_proof is not None and not crcat.lineage_proof.is_none():
                amount = uint128(amount + record.coin.amount)

        self.log.info(f"Pending approval balance for cat wallet {self.id()} is {amount}")
        return uint128(amount)

    async def convert_puzzle_hash(self, puzzle_hash: bytes32) -> bytes32:
        return puzzle_hash

    @staticmethod
    def get_metadata_from_record(coin_record: WalletCoinRecord) -> CRCATMetadata:
        metadata: MetadataTypes = coin_record.parsed_metadata()
        assert isinstance(metadata, CRCATMetadata)
        return metadata

    def coin_record_to_crcat(self, coin_record: WalletCoinRecord) -> CRCAT:
        if coin_record.coin_type not in {CoinType.CRCAT, CoinType.CRCAT_PENDING}:  # pragma: no cover
            raise ValueError(f"Attempting to spend a non-CRCAT coin: {coin_record.coin.name().hex()}")
        if coin_record.metadata is None:  # pragma: no cover
            raise ValueError(f"Attempting to spend a CRCAT coin without metadata: {coin_record.coin.name().hex()}")
        try:
            metadata: CRCATMetadata = CRCATWallet.get_metadata_from_record(coin_record)
            crcat: CRCAT = CRCAT(
                coin_record.coin,
                self.info.limitations_program_hash,
                metadata.lineage_proof,
                self.info.authorized_providers,
                self.info.proofs_checker.as_program(),
                (
                    construct_pending_approval_state(
                        metadata.inner_puzzle_hash, uint64(coin_record.coin.amount)
                    ).get_tree_hash()
                    if coin_record.coin_type == CoinType.CRCAT_PENDING
                    else metadata.inner_puzzle_hash
                ),
            )
            return crcat
        except Exception as e:  # pragma: no cover
            raise ValueError(f"Error parsing CRCAT metadata: {e}")

    async def get_lineage_proof_for_coin(self, coin: Coin) -> Optional[LineageProof]:  # pragma: no cover
        raise RuntimeError("get_lineage_proof_for_coin is a legacy method and is not available on CR-CAT wallets")

    async def _generate_unsigned_spendbundle(
        self,
        payments: list[Payment],
        action_scope: WalletActionScope,
        fee: uint64 = uint64(0),
        cat_discrepancy: Optional[tuple[int, Program, Program]] = None,  # (extra_delta, tail_reveal, tail_solution)
        coins: Optional[set[Coin]] = None,
        extra_conditions: tuple[Condition, ...] = tuple(),
        add_authorizations_to_cr_cats: bool = True,
    ) -> WalletSpendBundle:
        if cat_discrepancy is not None:
            extra_delta, tail_reveal, tail_solution = cat_discrepancy
        else:
            extra_delta, tail_reveal, tail_solution = 0, Program.to([]), Program.to([])
        payment_amount: int = sum(p.amount for p in payments)
        starting_amount: int = payment_amount - extra_delta
        if coins is None:
            cat_coins = list(
                await self.select_coins(
                    uint64(starting_amount),
                    action_scope,
                )
            )
        else:
            cat_coins = list(coins)

        cat_coins = sorted(cat_coins, key=Coin.name)  # need determinism because we need definitive origin coin

        selected_cat_amount = sum(c.amount for c in cat_coins)
        assert selected_cat_amount >= starting_amount

        # Figure out if we need to absorb/melt some XCK as part of this
        regular_chik_to_claim: int = 0
        if payment_amount > starting_amount:
            # TODO: The no coverage comment is because minting is broken for both this and the standard CAT wallet
            fee = uint64(fee + payment_amount - starting_amount)  # pragma: no cover
        elif payment_amount < starting_amount:
            regular_chik_to_claim = payment_amount

        need_chik_transaction = (fee > 0 or regular_chik_to_claim > 0) and (fee - regular_chik_to_claim != 0)

        # Calculate standard puzzle solutions
        change = selected_cat_amount - starting_amount
        primaries: list[Payment] = []
        for payment in payments:
            primaries.append(payment)

        if change > 0:
            origin_crcat_record = await self.wallet_state_manager.coin_store.get_coin_record(
                next(iter(cat_coins)).name()
            )
            if origin_crcat_record is None:
                raise RuntimeError("A CR-CAT coin was selected that we don't have a record for")  # pragma: no cover
            origin_crcat = self.coin_record_to_crcat(origin_crcat_record)
            if action_scope.config.tx_config.override(
                reuse_puzhash=(
                    True if not add_authorizations_to_cr_cats else action_scope.config.tx_config.reuse_puzhash
                )
            ).reuse_puzhash:
                change_puzhash = origin_crcat.inner_puzzle_hash
                for payment in payments:
                    if change_puzhash == payment.puzzle_hash and change == payment.amount:
                        # We cannot create two coins has same id, create a new puzhash for the change
                        change_puzhash = await self.standard_wallet.get_puzzle_hash(new=True)
                        break
            else:
                change_puzhash = await self.standard_wallet.get_puzzle_hash(new=True)
            primaries.append(Payment(change_puzhash, uint64(change), [change_puzhash]))

        # Find the VC Wallet
        vc_wallet: VCWallet
        for wallet in self.wallet_state_manager.wallets.values():
            if WalletType(wallet.type()) == WalletType.VC:
                assert isinstance(wallet, VCWallet)
                vc_wallet = wallet
                break
        else:
            raise RuntimeError("CR-CATs cannot be spent without an appropriate VC")  # pragma: no cover

        # Loop through the coins we've selected and gather the information we need to spend them
        vc: Optional[VerifiedCredential] = None
        vc_announcements_to_make: list[bytes] = []
        inner_spends: list[tuple[CRCAT, int, Program, Program]] = []
        first = True
        announcement: CreateCoinAnnouncement
        coin_ids: list[bytes32] = [coin.name() for coin in cat_coins]
        coin_records: list[WalletCoinRecord] = (
            await self.wallet_state_manager.coin_store.get_coin_records(coin_id_filter=HashFilter.include(coin_ids))
        ).records
        assert len(coin_records) == len(cat_coins)
        # sort the coin records to ensure they are in the same order as the CAT coins
        coin_records = [rec for rec in sorted(coin_records, key=lambda rec: coin_ids.index(rec.coin.name()))]
        for coin in coin_records:
            if vc is None:
                vc = await vc_wallet.get_vc_with_provider_in_and_proofs(
                    self.info.authorized_providers, self.info.proofs_checker.flags
                )

            if cat_discrepancy is not None:
                cat_condition = UnknownCondition(
                    opcode=Program.to(51),
                    args=[
                        Program.to(None),
                        Program.to(-113),
                        tail_reveal,
                        tail_solution,
                    ],
                )
                if first:
                    extra_conditions = (*extra_conditions, cat_condition)

            crcat: CRCAT = self.coin_record_to_crcat(coin)
            vc_announcements_to_make.append(crcat.expected_announcement())
            if first:
                announcement = CreateCoinAnnouncement(std_hash(b"".join([c.name() for c in cat_coins])), coin.name())
                if need_chik_transaction:
                    if fee > regular_chik_to_claim:
                        await self.create_tandem_xck_tx(
                            fee,
                            uint64(regular_chik_to_claim),
                            action_scope,
                            extra_conditions=(announcement.corresponding_assertion(),),
                        )
                        innersol = self.standard_wallet.make_solution(
                            primaries=primaries,
                            conditions=(*extra_conditions, announcement),
                        )
                    elif regular_chik_to_claim > fee:
                        xck_announcement = await self.create_tandem_xck_tx(
                            fee,
                            uint64(regular_chik_to_claim),
                            action_scope,
                        )
                        assert xck_announcement is not None
                        innersol = self.standard_wallet.make_solution(
                            primaries=primaries,
                            conditions=(*extra_conditions, xck_announcement, announcement),
                        )
                    else:
                        # TODO: what about when they are equal?
                        raise Exception("Equality not handled")
                else:
                    innersol = self.standard_wallet.make_solution(
                        primaries=primaries,
                        conditions=(*extra_conditions, announcement),
                    )
            else:
                innersol = self.standard_wallet.make_solution(
                    primaries=[],
                    conditions=(announcement.corresponding_assertion(),),
                )
            inner_derivation_record = (
                await self.wallet_state_manager.puzzle_store.get_derivation_record_for_puzzle_hash(
                    crcat.inner_puzzle_hash
                )
            )
            if inner_derivation_record is None:
                raise RuntimeError(  # pragma: no cover
                    f"CR-CAT {crcat} has an inner puzzle hash {crcat.inner_puzzle_hash} that we don't have the keys for"
                )
            inner_puzzle: Program = self.standard_wallet.puzzle_for_pk(inner_derivation_record.pubkey)
            inner_spends.append(
                (
                    crcat,
                    extra_delta if first else 0,
                    inner_puzzle,
                    innersol,
                )
            )
            first = False

        if vc is None:  # pragma: no cover
            raise RuntimeError("Spending no cat coins is not an appropriate use of _generate_unsigned_spendbundle")
        if vc.proof_hash is None:
            raise RuntimeError("CR-CATs found an appropriate VC but that VC contains no proofs")  # pragma: no cover

        proof_of_inclusions: Program = await vc_wallet.proof_of_inclusions_for_root_and_keys(
            vc.proof_hash, self.info.proofs_checker.flags
        )

        expected_announcements, coin_spends, _ = CRCAT.spend_many(
            inner_spends,
            proof_of_inclusions,
            Program.to(None),  # TODO: With more proofs checkers, this may need to be flexible. For now, it's hardcoded.
            vc.proof_provider,
            vc.launcher_id,
            vc.wrap_inner_with_backdoor().get_tree_hash() if add_authorizations_to_cr_cats else None,
        )
        if add_authorizations_to_cr_cats:
            await vc_wallet.generate_signed_transaction(
                vc.launcher_id,
                action_scope,
                extra_conditions=(
                    *expected_announcements,
                    announcement,
                    *(CreatePuzzleAnnouncement(ann) for ann in vc_announcements_to_make),
                ),
            )

        return WalletSpendBundle(coin_spends, G2Element())

    async def generate_signed_transaction(
        self,
        amounts: list[uint64],
        puzzle_hashes: list[bytes32],
        action_scope: WalletActionScope,
        fee: uint64 = uint64(0),
        coins: Optional[set[Coin]] = None,
        memos: Optional[list[list[bytes]]] = None,
        extra_conditions: tuple[Condition, ...] = tuple(),
        **kwargs: Unpack[GSTOptionalArgs],
    ) -> None:
        # (extra_delta, tail_reveal, tail_solution)
        cat_discrepancy: Optional[tuple[int, Program, Program]] = kwargs.get("cat_discrepancy", None)
        add_authorizations_to_cr_cats: bool = kwargs.get("add_authorizations_to_cr_cats", True)
        if memos is None:
            memos = [[] for _ in range(len(puzzle_hashes))]

        if not (len(memos) == len(puzzle_hashes) == len(amounts)):
            raise ValueError("Memos, puzzle_hashes, and amounts must have the same length")  # pragma: no cover

        payments = []
        for amount, puzhash, memo_list in zip(amounts, puzzle_hashes, memos):
            memos_with_hint: list[bytes] = [puzhash]
            memos_with_hint.extend(memo_list)
            # Force wrap the outgoing coins in the pending state if not going to us
            payments.append(
                Payment(
                    (
                        construct_pending_approval_state(puzhash, amount).get_tree_hash()
                        if puzhash != Offer.ph()
                        and not await self.wallet_state_manager.puzzle_store.puzzle_hash_exists(puzhash)
                        else puzhash
                    ),
                    amount,
                    memos_with_hint,
                )
            )

        spend_bundle = await self._generate_unsigned_spendbundle(
            payments,
            action_scope,
            fee,
            cat_discrepancy=cat_discrepancy,  # (extra_delta, tail_reveal, tail_solution)
            coins=coins,
            extra_conditions=extra_conditions,
            add_authorizations_to_cr_cats=add_authorizations_to_cr_cats,
        )

        async with action_scope.use() as interface:
            other_tx_removals: set[Coin] = {
                removal for tx in interface.side_effects.transactions for removal in tx.removals
            }
            other_tx_additions: set[Coin] = {
                addition for tx in interface.side_effects.transactions for addition in tx.additions
            }
            tx_list = [
                TransactionRecord(
                    confirmed_at_height=uint32(0),
                    created_at_time=uint64(int(time.time())),
                    to_puzzle_hash=payment.puzzle_hash,
                    amount=payment.amount,
                    fee_amount=fee,
                    confirmed=False,
                    sent=uint32(0),
                    spend_bundle=spend_bundle if i == 0 else None,
                    additions=list(set(spend_bundle.additions()) - other_tx_additions) if i == 0 else [],
                    removals=list(set(spend_bundle.removals()) - other_tx_removals) if i == 0 else [],
                    wallet_id=self.id(),
                    sent_to=[],
                    trade_id=None,
                    type=uint32(TransactionType.OUTGOING_TX.value),
                    name=spend_bundle.name() if i == 0 else payment.as_condition().get_tree_hash(),
                    memos=list(compute_memos(spend_bundle).items()),
                    valid_times=parse_timelock_info(extra_conditions),
                )
                for i, payment in enumerate(payments)
            ]

            interface.side_effects.transactions.extend(tx_list)

    async def claim_pending_approval_balance(
        self,
        min_amount_to_claim: uint64,
        action_scope: WalletActionScope,
        fee: uint64 = uint64(0),
        coins: Optional[set[Coin]] = None,
        min_coin_amount: Optional[uint64] = None,
        max_coin_amount: Optional[uint64] = None,
        excluded_coin_amounts: Optional[list[uint64]] = None,
        reuse_puzhash: Optional[bool] = None,
        extra_conditions: tuple[Condition, ...] = tuple(),
    ) -> None:
        # Select the relevant CR-CAT coins
        crcat_records: set[WalletCoinRecord] = await self.wallet_state_manager.coin_store.get_unspent_coins_for_wallet(
            self.id(), CoinType.CRCAT_PENDING
        )
        if coins is None:
            if max_coin_amount is None:
                max_coin_amount = uint64(self.wallet_state_manager.constants.MAX_COIN_AMOUNT)
            coins = await select_coins(
                await self.get_pending_approval_balance(),
                action_scope.config.tx_config.coin_selection_config,
                list(crcat_records),
                {},
                self.log,
                uint128(min_amount_to_claim),
            )

        # Select the relevant XCK coins
        if fee > 0:
            chik_coins = await self.standard_wallet.select_coins(
                fee,
                action_scope,
            )
        else:
            chik_coins = set()

        # Select the relevant VC coin
        vc_wallet: VCWallet = await self.wallet_state_manager.get_or_create_vc_wallet()
        vc: Optional[VerifiedCredential] = await vc_wallet.get_vc_with_provider_in_and_proofs(
            self.info.authorized_providers, self.info.proofs_checker.flags
        )
        if vc is None:  # pragma: no cover
            raise RuntimeError(f"No VC exists that can approve spends for CR-CAT wallet {self.id()}")
        if vc.proof_hash is None:
            raise RuntimeError(f"VC {vc.launcher_id} has no proofs to authorize transaction")  # pragma: no cover
        proof_of_inclusions: Program = await vc_wallet.proof_of_inclusions_for_root_and_keys(
            vc.proof_hash, self.info.proofs_checker.flags
        )

        # Generate the bundle nonce
        nonce: bytes32 = Program.to(
            [coin_as_list(c) for c in sorted(coins.union(chik_coins).union({vc.coin}), key=Coin.name)]
        ).get_tree_hash()

        # Make CR-CAT bundle
        crcats_and_puzhashes: list[tuple[CRCAT, bytes32]] = [
            (crcat, CRCATWallet.get_metadata_from_record(record).inner_puzzle_hash)
            for record in [r for r in crcat_records if r.coin in coins]
            for crcat in [self.coin_record_to_crcat(record)]
        ]
        expected_announcements, coin_spends, _ = CRCAT.spend_many(
            [
                (
                    crcat,
                    0,
                    construct_pending_approval_state(inner_puzhash, uint64(crcat.coin.amount)),
                    Program.to([nonce]),
                )
                for crcat, inner_puzhash in crcats_and_puzhashes
            ],
            proof_of_inclusions,
            Program.to(None),  # TODO: With more proofs checkers, this may need to be flexible. For now, it's hardcoded.
            vc.proof_provider,
            vc.launcher_id,
            vc.wrap_inner_with_backdoor().get_tree_hash(),
        )
        claim_bundle = WalletSpendBundle(coin_spends, G2Element())

        # Make the Fee TX
        if fee > 0:
            await self.create_tandem_xck_tx(
                fee,
                uint64(0),
                action_scope,
                extra_conditions=tuple(expected_announcements),
            )

        # Make the VC TX
        await vc_wallet.generate_signed_transaction(
            vc.launcher_id,
            action_scope,
            extra_conditions=(
                *extra_conditions,
                *expected_announcements,
                CreateCoinAnnouncement(nonce),
                *(CreatePuzzleAnnouncement(crcat.expected_announcement()) for crcat, _ in crcats_and_puzhashes),
            ),
        )

        async with action_scope.use() as interface:
            other_additions: set[Coin] = {rem for tx in interface.side_effects.transactions for rem in tx.additions}
            other_removals: set[Coin] = {rem for tx in interface.side_effects.transactions for rem in tx.removals}
            interface.side_effects.transactions.append(
                TransactionRecord(
                    confirmed_at_height=uint32(0),
                    created_at_time=uint64(int(time.time())),
                    to_puzzle_hash=await self.wallet_state_manager.main_wallet.get_puzzle_hash(False),
                    amount=uint64(sum(c.amount for c in coins)),
                    fee_amount=fee,
                    confirmed=False,
                    sent=uint32(0),
                    spend_bundle=claim_bundle,
                    additions=list(set(claim_bundle.additions()) - other_additions),
                    removals=list(set(claim_bundle.removals()) - other_removals),
                    wallet_id=self.id(),
                    sent_to=[],
                    trade_id=None,
                    type=uint32(TransactionType.INCOMING_TX.value),
                    name=claim_bundle.name(),
                    memos=list(compute_memos(claim_bundle).items()),
                    valid_times=parse_timelock_info(extra_conditions),
                )
            )

    async def match_puzzle_info(self, puzzle_driver: PuzzleInfo) -> bool:
        if (
            AssetType(puzzle_driver.type()) == AssetType.CAT
            and puzzle_driver["tail"] == self.info.limitations_program_hash
        ):
            inner_puzzle_driver: Optional[PuzzleInfo] = puzzle_driver.also()
            if inner_puzzle_driver is None:
                raise ValueError("Malformed puzzle driver passed to CRCATWallet.match_puzzle_info")  # pragma: no cover
            return (
                AssetType(inner_puzzle_driver.type()) == AssetType.CR
                and [bytes32(provider) for provider in inner_puzzle_driver["authorized_providers"]]
                == self.info.authorized_providers
                and ProofsChecker.from_program(uncurry_puzzle(inner_puzzle_driver["proofs_checker"]))
                == self.info.proofs_checker
            )
        return False

    async def get_puzzle_info(self, asset_id: bytes32) -> PuzzleInfo:
        return PuzzleInfo(
            {
                "type": AssetType.CAT.value,
                "tail": "0x" + self.info.limitations_program_hash.hex(),
                "also": {
                    "type": AssetType.CR.value,
                    "authorized_providers": ["0x" + provider.hex() for provider in self.info.authorized_providers],
                    "proofs_checker": self.info.proofs_checker.as_program(),
                },
            }
        )

    async def match_hinted_coin(self, coin: Coin, hint: bytes32) -> bool:
        """
        This matches coins that are either CRCATs with the hint as the inner puzzle, or CRCATs in the pending approval
        state that will come to us once claimed.
        """
        authorized_providers_hash: bytes32 = Program.to(self.info.authorized_providers).get_tree_hash()
        proofs_checker_hash: bytes32 = self.info.proofs_checker.as_program().get_tree_hash()
        hint_inner_hash: bytes32 = construct_cr_layer_hash(
            authorized_providers_hash,
            proofs_checker_hash,
            hint,
        )
        if (
            construct_cat_puzzle(
                Program.to(CAT_MOD_HASH),
                self.info.limitations_program_hash,
                hint_inner_hash,
                mod_code_hash=CAT_MOD_HASH_HASH,
            ).get_tree_hash_precalc(hint, CAT_MOD_HASH, CAT_MOD_HASH_HASH, hint_inner_hash)
            == coin.puzzle_hash
        ):
            return True

        pending_approval_inner_hash: bytes32 = construct_cr_layer_hash(
            authorized_providers_hash,
            proofs_checker_hash,
            construct_pending_approval_state(hint, uint64(coin.amount)).get_tree_hash(),
        )
        if (
            construct_cat_puzzle(
                Program.to(CAT_MOD_HASH),
                self.info.limitations_program_hash,
                pending_approval_inner_hash,
                mod_code_hash=CAT_MOD_HASH_HASH,
            ).get_tree_hash_precalc(CAT_MOD_HASH, CAT_MOD_HASH_HASH, pending_approval_inner_hash)
            == coin.puzzle_hash
        ):
            return True
        else:
            return False


if TYPE_CHECKING:
    _dummy: WalletProtocol[CATCoinData] = CRCATWallet()



from __future__ import annotations

import pytest

from chik._tests.util.db_connection import DBConnection
from chik.simulator.block_tools import BlockTools
from chik.types.full_block import FullBlock
from chik.types.header_block import HeaderBlock
from chik.wallet.key_val_store import KeyValStore


@pytest.mark.anyio
@pytest.mark.standard_block_tools
async def test_store(bt: BlockTools) -> None:
    async with DBConnection(1) as db_wrapper:
        store = await KeyValStore.create(db_wrapper)
        blocks = bt.get_consecutive_blocks(20)
        block: FullBlock = blocks[0]
        block_2: FullBlock = blocks[1]

        assert (await store.get_object("a", FullBlock)) is None
        await store.set_object("a", block)
        assert await store.get_object("a", FullBlock) == block
        await store.set_object("a", block)
        assert await store.get_object("a", FullBlock) == block
        await store.set_object("a", block_2)
        await store.set_object("a", block_2)
        assert await store.get_object("a", FullBlock) == block_2
        await store.remove_object("a")
        assert (await store.get_object("a", FullBlock)) is None

        for block in blocks:
            assert (await store.get_object(block.header_hash.hex(), FullBlock)) is None
            await store.set_object(block.header_hash.hex(), block)
            assert (await store.get_object(block.header_hash.hex(), FullBlock)) == block

        # Wrong type
        await store.set_object("a", block_2)
        with pytest.raises(Exception):
            await store.get_object("a", HeaderBlock)
